<?php

namespace Drupal\external_entities\FieldMapper;

use Drupal\Component\Plugin\ConfigurableInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\TypedData\DataDefinitionInterface;
use Drupal\Core\TypedData\DataReferenceInterface;
use Drupal\Core\TypedData\Plugin\DataType\DateTimeIso8601;
use Drupal\Core\TypedData\PrimitiveInterface;
use Drupal\Core\TypedData\Type\DateTimeInterface;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItem;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\external_entities\ExternalEntityInterface;
use Drupal\external_entities\ExternalEntityTypeInterface;
use Drupal\external_entities\Plugin\PluginFormTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A base class for field mappers.
 */
abstract class FieldMapperBase extends PluginBase implements FieldMapperInterface, ConfigurableInterface, PluginFormInterface {

  use PluginFormTrait;

  /**
   * Max length for mapping expressions.
   */
  public const MAPPING_FIELD_MAX_LENGTH = 2048;

  /**
   * Debug level of current instance.
   *
   * @var int
   */
  protected $debugLevel;

  /**
   * The external entity type this field mapper is configured for.
   *
   * @var \Drupal\external_entities\ExternalEntityTypeInterface
   */
  protected $externalEntityType;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The typed data manager.
   *
   * @var \Drupal\Core\TypedData\TypedDataManagerInterface
   */
  protected $typedDataManager;

  /**
   * List of entity mappable field definitions keyed by field name.
   *
   * @var \Drupal\Core\Field\FieldDefinitionInterface[]
   */
  protected $mappableFields;

  /**
   * Mapping definitions for all mappable fields keyed by field names.
   *
   * @var array
   */
  protected $fieldMappings;

  /**
   * Constructs a FieldMapperBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The identifier for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\TypedData\TypedDataManagerInterface $typed_data_manager
   *   The typed data manager.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFieldManagerInterface $entity_field_manager,
    TypedDataManagerInterface $typed_data_manager,
    TranslationInterface $string_translation
  ) {
    $configuration += $this->defaultConfiguration();
    if (!empty($configuration['_external_entity_type'])
        && $configuration['_external_entity_type'] instanceof ExternalEntityTypeInterface
    ) {
      $this->externalEntityType = $configuration['_external_entity_type'];
      unset($configuration['_external_entity_type']);
    }
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->typedDataManager = $typed_data_manager;
    $this->setStringTranslation($string_translation);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('typed_data_manager'),
      $container->get('string_translation')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel() :string {
    $plugin_definition = $this->getPluginDefinition();
    return $plugin_definition['label'];
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() :string {
    $plugin_definition = $this->getPluginDefinition();
    return $plugin_definition['description'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getDebugLevel() :int {
    return $this->debugLevel ?? 0;
  }

  /**
   * {@inheritdoc}
   */
  public function setDebugLevel(int $debug_level = 1) {
    $this->debugLevel = $debug_level;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMappings() :array {
    $configuration = $this->getConfiguration();
    return $configuration['field_mappings'] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapping(
    string $field_name,
    ?string $property_name = NULL
  ) :array|null {
    $field_mappings = $this->getFieldMappings();

    if (empty($field_mappings[$field_name])) {
      return NULL;
    }

    if (empty($property_name)) {
      return $field_mappings[$field_name];
    }

    return $field_mappings[$field_name][$property_name] ?? NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getMappedFieldName(
    string $field_name,
    ?string $property_name = NULL
  ) :?string {
    $source_id_field = NULL;
    $mapping_def = $this->getFieldMapping($field_name, $property_name);
    // Check if a property name was specified.
    if (!isset($property_name)) {
      // Nope, get the default property.
      if (1 == count($mapping_def)) {
        $mapping_def = current($mapping_def);
      }
      else {
        $mapping_def = [];
      }
    }
    if (!empty($mapping_def)
        && (FieldMapperInterface::MAPPING_DIRECT == $mapping_def['type'])
    ) {
      $source_id_field = $mapping_def['mapping'];
    }
    return $source_id_field;
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldMapping(
    string $field_name,
    ?string $property_name,
    string $mapping,
    int $mapping_type = FieldMapperInterface::MAPPING_DIRECT
  ) :array {
    if (!isset($property_name)) {
      // Get property name. Try from current mapping first.
      if (!empty($this->fieldMappings[$field_name])) {
        if (1 != count($this->fieldMappings[$field_name])) {
          // Property name required and missing, do nothing.
          return [];
        }
        $property_name = array_keys($this->fieldMappings[$field_name])[0];
      }
      else {
        // Try from field definition.
        $field_defs = $this->getMappableFields();
        if (empty($field_defs[$field_name])) {
          // No such field.
          return [];
        }
        $properties = $field_defs[$field_name]->getMappableFieldProperties();
        if (1 != count($properties)) {
          // More than one property, name required and missing, do nothing.
          return [];
        }
        $property_name = array_keys($properties)[0];
      }
    }
    $previous_mapping = $this->fieldMappings[$field_name][$property_name] ?? [];
    $this->fieldMappings[$field_name][$property_name] = [
      'type' => $mapping_type,
      'mapping' => $mapping,
    ];
    return $previous_mapping;
  }

  /**
   * Get the external entity type being operated for.
   *
   * @return \Drupal\external_entities\ExternalEntityTypeInterface
   *   The external entity type definition.
   */
  protected function getExternalEntityType() :ExternalEntityTypeInterface {
    return $this->externalEntityType;
  }

  /**
   * Get the fields that can be mapped.
   *
   * Computed fields are unmappable, which automatically excludes inherited
   * annotation fields as well. The annotation field is excluded as well.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface[]
   *   An associative array of field definitions, keyed by field name.
   */
  protected function getMappableFields() :array {
    if (!isset($this->mappableFields)) {
      $derived_entity_type = $this
        ->getExternalEntityType()
        ->getDerivedEntityType();

      $fields = [];
      if (!empty($derived_entity_type)) {
        $fields = $this
          ->entityFieldManager
          ->getFieldDefinitions(
            $derived_entity_type->id(),
            $derived_entity_type->id()
          );
      }
      $this->mappableFields = array_filter(
        $fields, function (FieldDefinitionInterface $field) {
          return $field->getName() !== ExternalEntityInterface::ANNOTATION_FIELD
            && !$field->isComputed();
        }
      );
    }

    return $this->mappableFields;
  }

  /**
   * Get the field properties that can be mapped.
   *
   * Field properties that are marked read-only (which include computed ones)
   * are considered unmappable.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition we want to extract mappable properties from.
   *
   * @return \Drupal\Core\TypedData\DataDefinitionInterface[]
   *   An array of property definitions of mappable properties, keyed by
   *   property name.
   */
  protected function getMappableFieldProperties(
    FieldDefinitionInterface $field_definition
  ) :array {
    $properties = $field_definition
      ->getFieldStorageDefinition()
      ->getPropertyDefinitions();
    return array_filter(
      $properties,
      function (DataDefinitionInterface $property) {
        $property_class = $property->getClass();
        return !$property->isReadOnly()
          && (is_subclass_of($property_class, PrimitiveInterface::class)
            || is_subclass_of($property_class, DataReferenceInterface::class));
      }
    );
  }

  /**
   * Get the fields for which a mapping is required.
   *
   * @return string[]
   *   An array of field names.
   */
  protected function getRequiredFieldMappings() :array {
    $fields = [
      'id',
      'uuid',
      'title',
    ];

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function extractEntityValuesFromRawData(array $raw_data) :array {
    $entity_values = [];

    $field_definitions = $this->getMappableFields();
    $context = [];

    foreach (array_keys($field_definitions) as $field_name) {
      $field_values = $this->extractFieldValuesFromRawData(
        $field_name,
        $raw_data,
        $context
      );
      if (!empty($field_values)) {
        $entity_values[$field_name] = $field_values;
      }
    }

    return $entity_values;
  }

  /**
   * {@inheritdoc}
   */
  public function extractFieldValuesFromRawData(
    string $field_name,
    array $raw_data,
    array &$context = []
  ) :?array {
    // Get field definition.
    $field_defs = $this->getMappableFields();
    if (empty($field_defs[$field_name])) {
      return NULL;
    }
    $field_definition = $field_defs[$field_name];

    $property_mappings = $this->getFieldMapping($field_name);
    if (empty($property_mappings)) {
      return NULL;
    }

    $properties_values = [];
    foreach ($property_mappings as $property_name => $mapping_def) {
      if (FieldMapperInterface::MAPPING_CONSTANT == $mapping_def['type']) {
        continue;
      }

      // Extract values.
      $value = $this->extractFieldPropertyValuesFromRawData(
        $field_name,
        $property_name,
        $raw_data,
        $context
      );

      // Make sure to not set references to empty arrays.
      if (($property_name === 'target_id')
        && (empty($value) || empty($value[0]))
      ) {
        continue;
      }
      $properties_values[$property_name] = $value;
    }

    // We have now collected all the mappable property values. Let's merge them
    // together so they represent a single field values array.
    $field_values = [];
    foreach ($properties_values as $property_name => $property_values) {
      foreach ($property_values as $delta => $property_value) {
        $field_values[$delta][$property_name] = $property_value;
      }
    }

    // Now that we have the field values, and the amount of deltas is known, we
    // can add in the constant mapped values.
    foreach ($property_mappings as $property_name => $mapping_def) {
      if (FieldMapperInterface::MAPPING_CONSTANT != $mapping_def['type']) {
        continue;
      }

      if (!empty($field_values)) {
        foreach ($field_values as &$field_value) {
          $field_value[$property_name] = $mapping_def['mapping'];
        }
      }
      else {
        $field_values = [
          [$property_name => $mapping_def['mapping']],
        ];
      }
    }

    // Depending on the property type, its value might need some more massaging.
    foreach ($field_values as &$properties_values) {
      foreach ($properties_values as $property_name => &$property_value) {
        $property_value = $this->processPropertyValue(
          $field_definition,
          $property_name,
          $property_value
        );
      }
    }

    return $field_values;
  }

  /**
   * Processes a property value.
   *
   * Provides conversions for special data types and makes sure a property is in
   * the correct PHP value as expected by the data type.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   Definition of the field we are extracting for.
   * @param string $property_name
   *   Name of the property we are extracting.
   * @param mixed $property_value
   *   The value that needs to be processed.
   *
   * @return mixed
   *   The processed value.
   */
  protected function processPropertyValue(
    FieldDefinitionInterface $field_definition,
    string $property_name,
    $property_value
  ) {
    // Create the typed data instance.
    $property_definition = $field_definition
      ->getFieldStorageDefinition()
      ->getPropertyDefinition($property_name);
    $typed_data = $this->typedDataManager->create($property_definition);

    // Provide rudimentary support for datetime-based fields by making sure
    // they are in the format as expected by Drupal.
    if (in_array(DateTimeInterface::class, class_implements($typed_data, TRUE))) {
      $timestamp = $property_value !== NULL && !is_numeric($property_value)
        ? strtotime($property_value)
        : $property_value;

      if (is_numeric($timestamp)) {
        if (get_class($typed_data) === 'Drupal\Core\TypedData\Plugin\DataType\DateTimeIso8601') {
          assert($typed_data instanceof DateTimeIso8601);
          $datetime_type = $field_definition
            ->getFieldStorageDefinition()
            ->getSetting('datetime_type');

          if ($datetime_type === DateTimeItem::DATETIME_TYPE_DATE) {
            $storage_format = DateTimeItemInterface::DATE_STORAGE_FORMAT;
          }
          else {
            $storage_format = DateTimeItemInterface::DATETIME_STORAGE_FORMAT;
          }

          // Use setValue so timezone is not set.
          $typed_data->setValue(gmdate($storage_format, $timestamp));
        }
        else {
          $typed_data->setDateTime(
            DrupalDateTime::createFromTimestamp($timestamp)
          );
        }
      }
    }
    else {
      $typed_data->setValue($property_value);
    }

    // Convert the property value to the correct PHP type as expected by this
    // specific property type.
    if ($typed_data instanceof PrimitiveInterface) {
      $property_value = $typed_data->getCastedValue();
    }

    return $property_value;
  }

  /**
   * {@inheritdoc}
   */
  public function extractIdFromRawData(array $raw_data) :string|int|null {
    $id_field_values = $this->extractFieldPropertyValuesFromRawData(
      'id',
      'value',
      $raw_data
    );
    return reset($id_field_values);
  }

  /**
   * {@inheritdoc}
   */
  public function createRawDataFromEntityValues(
    array $entity_values,
    array $original_values
  ) :array {
    $raw_data = $original_values;

    $mappable_fields = $this->getMappableFields();
    $context = [];

    foreach ($entity_values as $field_name => $field_values) {
      $this->addFieldValuesToRawData(
        $mappable_fields[$field_name],
        $field_values,
        $raw_data,
        $context
      );
    }

    return $raw_data;
  }

  /**
   * Adds field values to a raw data array for a given field.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition for the field being mapped.
   * @param array $field_values
   *   The field values to add to the raw data.
   * @param array &$raw_data
   *   The raw data array being built/completed.
   * @param array &$context
   *   Any contextual data that needs to be maintained during the whole
   *   creation process for raw data.
   */
  abstract protected function addFieldValuesToRawData(
    FieldDefinitionInterface $field_definition,
    array $field_values,
    array &$raw_data,
    array &$context
  );

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = $configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [];
  }

  /**
   * Form constructor.
   *
   * @param array $form
   *   An associative array containing the initial structure of the plugin form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form. Calling code should pass on a subform
   *   state created through
   *   \Drupal\Core\Form\SubformState::createForSubform().
   *
   * @return array
   *   The form structure.
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form['#attached']['library'][] = 'external_entities/external_entities';

    $form['help'] = [
      '#type' => 'container',
    ];

    $form['field_mappings'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Field mappings'),
    ];

    $mappable_fields = $this->getMappableFields();
    foreach ($mappable_fields as $field_name => $field_definition) {
      $form['field_mappings'][$field_name] = $this->buildFieldMappingElement($field_definition);
    }

    return $form;
  }

  /**
   * Build a form element for configuring a field mapping.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   *
   * @return array
   *   The form element for configuring the field.
   */
  protected function buildFieldMappingElement(
    FieldDefinitionInterface $field_definition
  ) :array {
    $element = [];

    $property_definitions = $this->getMappableFieldProperties($field_definition);
    // Regroup multiple properties in a fieldset.
    if (count($property_definitions) > 1) {
      $element = [
        '#type' => 'details',
        '#title' => $this->t(
          '@field field',
          ['@field' => $field_definition->getLabel()]
        ),
        '#open' => TRUE,
        '#attributes' => ['class' => ['xntt-field-details']],
      ];
    }

    foreach ($property_definitions as $property_name => $property_definition) {
      $element[$property_name] = $this->buildFieldPropertyMappingElement(
        $field_definition,
        $property_name,
        $property_definition
      );
    }

    return $element;
  }

  /**
   * Build a form element for configuring a field property mapping.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   * @param string $property_name
   *   The property name.
   * @param \Drupal\Core\TypedData\DataDefinitionInterface $property_definition
   *   The property definition.
   *
   * @return array
   *   A renderable form element.
   */
  protected function buildFieldPropertyMappingElement(
    FieldDefinitionInterface $field_definition,
    string $property_name,
    DataDefinitionInterface $property_definition
  ) :array {
    $mapping_def = $this->getFieldMapping(
      $field_definition->getName(),
      $property_name
    );

    $property_definitions = $this->getMappableFieldProperties($field_definition);
    // Choose title between field label and property label.
    if (count($property_definitions) > 1) {
      // Property.
      $label = $property_definition->getLabel();
    }
    else {
      // Field.
      $label = $field_definition->getLabel();
    }

    $is_required_field = in_array(
      $field_definition->getName(),
      $this->getRequiredFieldMappings(),
      TRUE
    );
    $is_main_property = $field_definition->getFieldStorageDefinition()->getMainPropertyName() === $property_name;

    $element = [
      '#type' => 'container',
      '#attributes' => ['class' => ['xntt-inline']],
    ];
    $element['mapping'] = [
      '#title' => $label,
      '#type' => 'textfield',
      '#default_value' => $mapping_def['mapping'] ?? '',
      '#required' => $is_required_field && $is_main_property,
      '#maxlength' => static::MAPPING_FIELD_MAX_LENGTH,
      '#attributes' => ['class' => ['xntt-mapping']],
      '#label_attributes' => ['class' => ['xntt-mapping-label']],
    ];
    $element['type'] = [
      '#type' => 'select',
      '#title' => $this->t('Type'),
      '#default_value' =>
      $mapping_def['type']
      ?? FieldMapperInterface::MAPPING_DIRECT,
      '#options' => $this->getPossibleFieldPropertyMappingTypes(
        $field_definition,
        $property_name,
        $property_definition
      ),
      '#attributes' => ['class' => ['xntt-type']],
      '#label_attributes' => ['class' => ['xntt-type-label']],
    ];
    if ('id' == $field_definition->getName()) {
      $element['type']['#attributes']['readonly'] = TRUE;
    }

    return $element;
  }

  /**
   * Returns the list of possible mapping types for the given field property.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   * @param string $property_name
   *   The property name.
   * @param \Drupal\Core\TypedData\DataDefinitionInterface $property_definition
   *   The property definition.
   *
   * @return array
   *   An array of possible mapping types as a set of options for a 'select'
   *   form element.
   */
  protected function getPossibleFieldPropertyMappingTypes(
    FieldDefinitionInterface $field_definition,
    string $property_name,
    DataDefinitionInterface $property_definition
  ) :array {
    $possible_types = [
      FieldMapperInterface::MAPPING_DIRECT => $this->t('Field name'),
      FieldMapperInterface::MAPPING_CONSTANT => $this->t('Constant'),
    ];

    if ('id' == $field_definition->getName()) {
      $possible_types = [
        FieldMapperInterface::MAPPING_DIRECT => $this->t('Field name'),
      ];
    }
    return $possible_types;
  }

  /**
   * Form validation handler.
   *
   * @param array $form
   *   An associative array containing the structure of the plugin form as built
   *   by static::buildConfigurationForm().
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form. Calling code should pass on a subform
   *   state created through
   *   \Drupal\Core\Form\SubformState::createForSubform().
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    // Make sure required field mappings are not mapped to a constant value.
    $mappable_fields = $this->getMappableFields();
    if (empty($mappable_fields)) {
      return;
    }

    foreach ($this->getRequiredFieldMappings() as $field_name) {
      if (!array_key_exists($field_name, $mappable_fields)) {
        continue;
      }

      $main_property_name = $mappable_fields[$field_name]
        ->getFieldStorageDefinition()
        ->getMainPropertyName();
      $mapping = $form_state->getValue([
        'field_mappings',
        $field_name,
        $main_property_name,
        'mapping',
      ]);
      $mapping_type = $form_state->getValue([
        'field_mappings',
        $field_name,
        $main_property_name,
        'type',
      ]);
      if (FieldMapperInterface::MAPPING_CONSTANT == $mapping_type) {
        $form_state->setErrorByName($field_name, $this->t(
          'The @field_name field cannot be mapped to a constant value.',
          [
            '@field_name' => $mappable_fields[$field_name]->getLabel(),
          ]
        ));
      }
    }
  }

  /**
   * Form submit handler.
   *
   * @param array $form
   *   An associative array containing the structure of the plugin form as built
   *   by static::buildConfigurationForm().
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form. Calling code should pass on a subform
   *   state created through
   *   \Drupal\Core\Form\SubformState::createForSubform().
   */
  public function submitConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    // Remove empty mappings.
    $field_mappings = $form_state->getValue(['field_mappings'], []);
    $filtered_field_mappings = NestedArray::filter($field_mappings);
    $filtered_field_mappings = array_filter($filtered_field_mappings);
    $form_state->setValue(['field_mappings'], $filtered_field_mappings);
  }

}
