<?php

namespace Drupal\external_entities\FieldMapper;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;

/**
 * Interface for field mapper plugins.
 *
 * Field mappers control how raw data is mapped into and out of entity objects.
 */
interface FieldMapperInterface extends PluginInspectionInterface, ContainerFactoryPluginInterface {

  /**
   * Mapping type that maps a constant value.
   */
  const MAPPING_CONSTANT = 1;

  /**
   * Mapping type that maps a raw source field value to a Drupal field.
   */
  const MAPPING_DIRECT = 2;

  /**
   * Mapping type that maps a source field using the simple field mapper syntax.
   */
  const MAPPING_SIMPLE = 3;

  /**
   * Mapping type that maps a source field to a Drupal field with an expression.
   */
  const MAPPING_EXPRESSION = 4;

  /**
   * Mapping type that maps an already mapped Drupal field value to another.
   */
  const MAPPING_ALIAS = 5;

  /**
   * Returns the administrative label for this field mapper plugin.
   *
   * @return string
   *   The field mappers administrative label.
   */
  public function getLabel() :string;

  /**
   * Returns the administrative description for this field mapper plugin.
   *
   * @return string
   *   The field mappers administrative description.
   */
  public function getDescription() :string;

  /**
   * Returns debug level for current instance.
   *
   * @return int
   *   The debug level. 0 means debugging is off.
   */
  public function getDebugLevel() :int;

  /**
   * Sets debug level for current instance.
   *
   * @param int $debug_level
   *   Sets the debug level. A non-0 value means debugging is enabled while 0
   *   disables  debugging. Default to 1 (enabled).
   */
  public function setDebugLevel(int $debug_level = 1);

  /**
   * Gets all configured field mappings.
   *
   * @return array
   *   An associative array of field mappings, keyed by the field name. Each
   *   value is an associative array of mapping definitions, keyed by the
   *   property name. See self::getFieldMapping() for mapping definition
   *   structure.
   */
  public function getFieldMappings() :array;

  /**
   * Gets the mapping for a field.
   *
   * @param string $field_name
   *   The field name.
   * @param string|null $property_name
   *   (optional) The property name.
   *
   * @return array|null
   *   If called without a property name, all the property mapping definitions
   *   of the field are returned as an associative array of mapping definitions,
   *   keyed by the property name. If called with a property name, the mapping
   *   definition of that property is returned. NULL is returned if no mapping
   *   could be found for the given field and/or property.
   *   A mapping definition is an array with at least 2 keys: 'type' which is
   *   one of the constants FieldMapperInterface::MAPPING_* and 'mapping' which
   *   contains the mapping expression.
   */
  public function getFieldMapping(
    string $field_name,
    ?string $property_name = NULL
  ) :array|null;

  /**
   * Sets the given mapping for a given field and property.
   *
   * @param string $field_name
   *   The field name.
   * @param string|null $property_name
   *   The property name is required for fields with multiple properties. Null
   *   can be used for fields with a single property. Otherwise, no mapping will
   *   be set.
   * @param string $mapping
   *   Mapping expression.
   * @param int $mapping_type
   *   Mapping type. One of the MAPPING_* constants.
   *
   * @return array
   *   Returns previous mapping if one or an empty array if none.
   */
  public function setFieldMapping(
    string $field_name,
    ?string $property_name,
    string $mapping,
    int $mapping_type = FieldMapperInterface::MAPPING_DIRECT
  ) :array;

  /**
   * Returns the source field name mapped to the given Drupal field property.
   *
   * @param string $field_name
   *   The field name.
   * @param string|null $property_name
   *   The property name is required for fields with multiple properties. Null
   *   can be used for fields with a single property. Otherwise, NULL will be
   *   returned.
   *
   * @return string|null
   *   The source field name used for the givend Drupal field or NULL if not
   *   available.
   */
  public function getMappedFieldName(
    string $field_name,
    ?string $property_name = NULL
  ) :?string;

  /**
   * Extracts entity values from raw data.
   *
   * @param array $raw_data
   *   The raw data as received by the external service.
   *
   * @return array
   *   An array of values to set on the external entity object, keyed by
   *   property name. See an example of the structure below. This is directly
   *   passed on to the entity object constructor.
   *
   * @see \Drupal\Core\Entity\EntityBase::__construct()
   *
   * @code
   *   <?php
   *     [
   *       'field1' => [
   *         0 => [
   *           'property1' => ...,
   *           'property2' => ...,
   *           ...
   *         ],
   *         1 => [
   *           'property1' => ...,
   *           'property2' => ...,
   *           ...
   *         ],
   *     ],
   *     [
   *       'field2' => [
   *         0 => [
   *           'property1' => ...,
   *           'property2' => ...,
   *           ...
   *         ],
   *         1 => [
   *           'property1' => ...,
   *           'property2' => ...,
   *           ...
   *         ],
   *     ],
   *   ?>
   * @endcode
   */
  public function extractEntityValuesFromRawData(array $raw_data) :array;

  /**
   * Extracts field values from raw data for a given field.
   *
   * @param string $field_name
   *   Name of the field we are extracting for.
   * @param array $raw_data
   *   The raw data to extract the field values from.
   * @param array &$context
   *   Any contextual data that needs to be maintained during the whole
   *   extraction process for an external entity.
   *
   * @return array|null
   *   An array of field values, or NULL if none.
   */
  public function extractFieldValuesFromRawData(
    string $field_name,
    array $raw_data,
    array &$context = []
  ) :?array;

  /**
   * Extracts field property values from raw data for a given field property.
   *
   * @param string $field_name
   *   Name of the field we are extracting for.
   * @param string $property_name
   *   Name of the property we are extracting.
   * @param array $raw_data
   *   The raw data to extract the property values from.
   * @param array &$context
   *   Any contextual data that needs to be maintained during the whole
   *   extraction process for an external entity.
   *
   * @return array
   *   An array of property values, empty array if none.
   */
  public function extractFieldPropertyValuesFromRawData(
    string $field_name,
    string $property_name,
    array $raw_data,
    array &$context = []
  ) :array;

  /**
   * Extract the unique ID from the raw data.
   *
   * @param array $raw_data
   *   The raw data from the external service.
   *
   * @return string|int|null
   *   The unique ID the item is represented by or NULL if none could be found.
   */
  public function extractIdFromRawData(array $raw_data) :string|int|null;

  /**
   * Creates raw data from the given entity values.
   *
   * This method can be used both to generate raw data from all entity values or
   * for a single field or field property value depending on the $entity_values
   * parameter content.
   * Original values are needed to regenerate the full raw data set when some
   * raw data fields are not mapped and therefore not in the $entity_values
   * while they should remain saved on the source side even if they were not
   * used as they might be used by other processes.
   *
   * @param array $entity_values
   *   An associative array of entity values to be mapped to raw data. This
   *   value must have a structure similar to the one below.
   *   @code
   *     <?php
   *       [
   *         'field1' => [
   *           0 => [
   *             'property1' => ...,
   *             'property2' => ...,
   *             ...
   *           ],
   *           1 => [
   *             'property1' => ...,
   *             'property2' => ...,
   *             ...
   *           ],
   *       ],
   *       [
   *         'field2' => [
   *           0 => [
   *             'property1' => ...,
   *             'property2' => ...,
   *             ...
   *           ],
   *           1 => [
   *             'property1' => ...,
   *             'property2' => ...,
   *             ...
   *           ],
   *       ],
   *     ?>
   *   @endcode
   * @param array $original_values
   *   A source array of data that will be overrided by the $entity_values.
   *
   * @return array
   *   The raw data ready to be sent to the external service.
   */
  public function createRawDataFromEntityValues(
    array $entity_values,
    array $original_values
  ) :array;

}
