<?php

namespace Drupal\external_entities\Plugin\ExternalEntities\FieldMapper;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\TypedData\DataDefinitionInterface;
use Drupal\Core\Url;
use Drupal\external_entities\FieldMapper\FieldMapperBase;
use Drupal\external_entities\FieldMapper\FieldMapperInterface;
use JsonPath\JsonObject;

/**
 * A field mapper that uses JSONPath expressions.
 *
 * Multi-valued fields should be mapped using JSONPath expressions that result
 * in an array of values being returned.
 *
 * @FieldMapper(
 *   id = "jsonpath",
 *   label = @Translation("JSONPath"),
 *   description = @Translation("Maps fields based on JSONPath expressions.")
 * )
 */
class JsonPath extends FieldMapperBase {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form += parent::buildConfigurationForm($form, $form_state);

    $links = [
      Link::fromTextAndUrl(
        $this->t('JSONPath syntax and limitation'),
        Url::fromUri(
          'https://github.com/Galbar/JsonPath-PHP#jsonpath-language'
        )
      ),

      Link::fromTextAndUrl(
        $this->t('JSONPath examples'),
        Url::fromUri(
          'https://support.smartbear.com/alertsite/docs/monitors/api/endpoint/jsonpath.html'
        )
      ),
    ];

    $form['help']['jsonpath'] = [
      '#theme' => 'item_list',
      '#type' => 'ul',
      '#prefix' => '<p>' . $this->t('See these documentation links:') . '</p>',
      '#items' => array_map(function (Link $link) {
        return $link->toRenderable();
      }, $links),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function getPossibleFieldPropertyMappingTypes(
    FieldDefinitionInterface $field_definition,
    string $property_name,
    DataDefinitionInterface $property_definition
  ) :array {
    $possible_types = [
      FieldMapperInterface::MAPPING_EXPRESSION => $this->t('JSonPath'),
      FieldMapperInterface::MAPPING_DIRECT => $this->t('Field name'),
      FieldMapperInterface::MAPPING_CONSTANT => $this->t('Constant'),
    ];

    if ('id' == $field_definition->getName()) {
      $possible_types = [
        FieldMapperInterface::MAPPING_DIRECT => $this->t('Field name'),
      ];
    }
    return $possible_types;
  }

  /**
   * {@inheritdoc}
   */
  public function getMappedFieldName(
    string $field_name,
    ?string $property_name = NULL
  ) :?string {
    $source_id_field = NULL;
    $mapping_def = $this->getFieldMapping($field_name, $property_name);
    // Check if a property name was specified.
    if (!isset($property_name)) {
      // Nope, get the default property.
      if (1 == count($mapping_def)) {
        $mapping_def = current($mapping_def);
      }
      else {
        $mapping_def = [];
      }
    }
    if (!empty($mapping_def)) {
      if (FieldMapperInterface::MAPPING_DIRECT == $mapping_def['type']) {
        $source_id_field = $mapping_def['mapping'];
      }
      elseif ((FieldMapperInterface::MAPPING_EXPRESSION == $mapping_def['type'])
        && (preg_match(
              '#^\$(?:\.(\w+)|\[[\'"](\w+)["\']\])$#',
              $mapping_def['mapping'],
              $matches,
              PREG_UNMATCHED_AS_NULL
            )
        )
      ) {
        $source_id_field = $matches[1] ?? $matches[2];
      }
    }
    return $source_id_field;
  }

  /**
   * {@inheritdoc}
   */
  public function extractFieldPropertyValuesFromRawData(
    string $field_name,
    string $property_name,
    array $raw_data,
    array &$context = []
  ) :array {
    $mapping = $this->getFieldMapping(
      $field_name,
      $property_name
    );
    if (empty($mapping)) {
      return [];
    }

    if (FieldMapperInterface::MAPPING_DIRECT == $mapping['type']) {
      return $raw_data[$mapping['mapping']];
    }

    if (empty($context['jsonpath_object'])) {
      $context['jsonpath_object'] = new JsonObject($raw_data);
    }

    return $context['jsonpath_object']->get($mapping['mapping']) ?: [];
  }

  /**
   * {@inheritdoc}
   */
  protected function addFieldValuesToRawData(
    FieldDefinitionInterface $field_definition,
    array $field_values,
    array &$raw_data,
    array &$context
  ) {
    if (empty($context['jsonpath_object'])) {
      $context['jsonpath_object'] = new JsonObject($raw_data);
    }

    $property_mappings = $this->getFieldMapping($field_definition->getName());
    if (empty($property_mappings)) {
      return;
    }

    $field_cardinality = $field_definition
      ->getFieldStorageDefinition()
      ->getCardinality();

    // Convert [delta][property] structure to [property][delta] structure, so
    // that each property can be set in the raw data all at once in one setter
    // operation.
    $property_values = [];
    foreach ($field_values as $delta => $field_value) {
      foreach ($field_value as $property_name => $property_value) {
        $property_values[$property_name][$delta] = $property_value;
      }
    }

    foreach ($property_mappings as $property_name => $mapping_def) {
      // Skip constant values.
      if (FieldMapperInterface::MAPPING_CONSTANT == $mapping_def['type']) {
        continue;
      }

      $json_value = $property_values[$property_name] ?? [];
      // Unwrap the singleton array if the cardinality of the field is 1.
      if ($field_cardinality === 1) {
        $json_value = array_shift($json_value);
      }

      if (!empty($json_value)) {
        $context['jsonpath_object']->set($mapping['mapping'], $json_value);
      }
    }

    // $json_object carries state between the fields, so this is just
    // reconciling the output of mapping with the state building up in the
    // JSONPath interpreter.
    $raw_data = $context['jsonpath_object']->getValue();
  }

}
