<?php

namespace Drupal\external_entities\Plugin\ExternalEntities\FieldMapper;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\TypedData\DataDefinitionInterface;
use Drupal\external_entities\FieldMapper\FieldMapperBase;
use Drupal\external_entities\FieldMapper\FieldMapperInterface;

/**
 * This is a short description of the function.
 *
 * This is additional information about the function.
 * It should be on a separate line from the short description.
 *
 * @param int $parameter Description of the parameter.
 *
 * @return string Description of the return value.
 *
 * @FieldMapper(
 *   id = "simple",
 *   label = @Translation("Simple"),
 *   description = @Translation("Maps entity fields to raw data using simple
 *   path expressions.")
 * )
 *
 * @package Drupal\external_entities\Plugin\ExternalEntities\FieldMapper
 */
class SimpleFieldMapper extends FieldMapperBase {

  /**
   * {@inheritdoc}
   */
  protected function getRequiredFieldMappings() :array {
    return [
      'id',
      'title',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form += parent::buildConfigurationForm($form, $form_state);

    $form['field_mappings']['uuid']['value']['#description'] = $this->t(
      'It is advised to map this field'
    );
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function getPossibleFieldPropertyMappingTypes(
    FieldDefinitionInterface $field_definition,
    string $property_name,
    DataDefinitionInterface $property_definition
  ) :array {
    $possible_types = [
      FieldMapperInterface::MAPPING_DIRECT => $this->t('Field name'),
      FieldMapperInterface::MAPPING_SIMPLE => $this->t('Simple mapping'),
      FieldMapperInterface::MAPPING_CONSTANT => $this->t('Constant'),
    ];

    if ('id' == $field_definition->getName()) {
      $possible_types = [
        FieldMapperInterface::MAPPING_DIRECT => $this->t('Field name'),
      ];
    }
    return $possible_types;
  }

  /**
   * {@inheritdoc}
   */
  public function getMappedFieldName(
    string $field_name,
    ?string $property_name = NULL
  ) :?string {
    $source_id_field = NULL;
    $mapping_def = $this->getFieldMapping($field_name, $property_name);
    // Check if a property name was specified.
    if (!isset($property_name)) {
      // Nope, get the default property.
      if (1 == count($mapping_def)) {
        $mapping_def = current($mapping_def);
      }
      else {
        $mapping_def = [];
      }
    }
    if (!empty($mapping_def)) {
      if (FieldMapperInterface::MAPPING_DIRECT == $mapping_def['type']) {
        $source_id_field = $mapping_def['mapping'];
      }
      elseif ((FieldMapperInterface::MAPPING_SIMPLE == $mapping_def['type'])
        && (preg_match('/^([\w#@$\-]+)$/', $mapping_def['mapping'], $matches))
      ) {
        $source_id_field = $matches[1];
      }
    }
    return $source_id_field;
  }

  /**
   * {@inheritdoc}
   */
  public function extractFieldPropertyValuesFromRawData(
    string $field_name,
    string $property_name,
    array $raw_data,
    array &$context = []
  ) :array {
    $mapping = $this->getFieldMapping(
      $field_name,
      $property_name
    );

    if (empty($mapping)
      || (!in_array(
            $mapping['type'],
            [
              FieldMapperInterface::MAPPING_DIRECT,
              FieldMapperInterface::MAPPING_SIMPLE,
            ]
          )
      )
    ) {
      return [];
    }
    // Works with both MAPPING_DIRECT and MAPPING_SIMPLE.
    $mapping_keys = explode('/', $mapping['mapping']);

    return $this->recursiveMapFieldFromRawData(
      $raw_data,
      $mapping_keys,
      [],
      0,
      $property_name
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function addFieldValuesToRawData(
    FieldDefinitionInterface $field_definition,
    array $field_values,
    array &$raw_data,
    array &$context
  ) {

    $property_mappings = $this->getFieldMapping($field_definition->getName());
    if (empty($property_mappings)) {
      return;
    }

    // Convert [delta][property] structure to [property][delta] structure, so
    // that each property can be set in the raw data all at once in one setter
    // operation.
    $property_values = [];
    foreach ($field_values as $delta => $field_value) {
      foreach ($field_value as $property_name => $property_value) {
        $property_values[$property_name][$delta] = $property_value;
      }
    }
    foreach ($property_mappings as $property_name => $mapping_def) {
      // Skip constant values.
      if (FieldMapperInterface::MAPPING_CONSTANT == $mapping_def['type']) {
        continue;
      }

      $data_value = $property_values[$property_name] ?? [];

      $delta_index = 0;
      if (!empty($data_value)) {
        foreach ($data_value as $value) {
          $qualified_mapping = str_replace('*', $delta_index, $mapping_def['mapping']);
          $mapping_keys = explode('/', $qualified_mapping);
          NestedArray::setValue($raw_data, $mapping_keys, $value);
          $delta_index++;
        }
        // Clear remaining extra items if some.
        if (FALSE !== strpos($mapping_def['mapping'], '*')) {
          do {
            $qualified_mapping = str_replace('*', $delta_index, $mapping_def['mapping']);
            $mapping_keys = explode('/', $qualified_mapping);
            NestedArray::unsetValue($raw_data, $mapping_keys, $unset);
            $delta_index++;
          } while ($unset);
        }
      }
    }
  }

  /**
   * Populates the values of a field raw by recursively navigating mapping keys.
   *
   * @param array $raw_data
   *   The raw values to map into the output.
   * @param array $remaining_mapping_keys
   *   All the mapping keys that remain to be examined recursively.
   * @param array $seen_mapping_keys
   *   All the mapping keys that have been examined so far.
   * @param int $field_delta
   *   The current delta within the field being populated.
   * @param string $field_property_name
   *   The machine name of the field property being populated.
   *
   * @return array
   *   The list of property values or an empty array.
   */
  protected function recursiveMapFieldFromRawData(
    array $raw_data,
    array $remaining_mapping_keys,
    array $seen_mapping_keys,
    int $field_delta,
    string $field_property_name
  ): array {
    $current_mapping_key = array_shift($remaining_mapping_keys);

    // Case 1: End of recursion -- set the field property value.
    if ($current_mapping_key === NULL) {
      return $this->extractRawData(
        $raw_data,
        $seen_mapping_keys,
        $field_delta,
        $field_property_name
      );
    }
    // Case 2: Iterate and recurse over a list of values.
    elseif ($current_mapping_key == '*') {
      return $this->recursiveMapListFieldFromRawData(
        $raw_data,
        $remaining_mapping_keys,
        $seen_mapping_keys,
        $field_property_name
      );
    }
    // Case 3: Recurse into a single-valued key of the raw data.
    else {
      $new_seen_mapping_keys =
        array_merge($seen_mapping_keys, [$current_mapping_key]);

      return $this->recursiveMapFieldFromRawData(
        $raw_data,
        $remaining_mapping_keys,
        $new_seen_mapping_keys,
        $field_delta,
        $field_property_name
      );
    }
  }

  /**
   * Populates the field property based on the provided keys.
   *
   * This function takes the provided keys as input and
   * populates the corresponding field property.
   *
   * @param array $raw_data
   *   The raw values from which to obtain the field value.
   * @param array $raw_keys
   *   The nested array keys of the raw data that designate the location in the
   *   raw data where the value is to be extracted.
   * @param int $field_delta
   *   The current delta within the field being populated.
   * @param string $field_property_name
   *   The machine name of the field property being populated.
   *
   * @return array
   *   List of property values.
   */
  protected function extractRawData(
    array $raw_data,
    array $raw_keys,
    int $field_delta,
    string $field_property_name
  ): array {
    $property_value = NestedArray::getValue($raw_data, $raw_keys);
    return is_array($property_value) ? $property_value : [$property_value];
  }

  /**
   * Extracts a list of values from the raw data and populates the equivalent.
   *
   * @param array $raw_data
   *   The raw values from which to obtain the list values.
   * @param array $remaining_mapping_keys
   *   All the mapping keys that remain to be examined recursively.
   * @param array $seen_mapping_keys
   *   All the mapping keys that have been examined so far.
   * @param string $field_property_name
   *   The machine name of the field property being populated.
   */
  protected function recursiveMapListFieldFromRawData(
    array $raw_data,
    array $remaining_mapping_keys,
    array $seen_mapping_keys,
    string $field_property_name
  ): array {
    $raw_values = NestedArray::getValue($raw_data, $seen_mapping_keys);
    $field_values = [];

    if (!is_array($raw_values)) {
      $raw_values = [$raw_values];
    }

    for ($delta_index = 0; $delta_index < count($raw_values); ++$delta_index) {
      $new_seen_mapping_keys = array_merge($seen_mapping_keys, [$delta_index]);

      $values = $this->recursiveMapFieldFromRawData(
        $raw_data,
        $remaining_mapping_keys,
        $new_seen_mapping_keys,
        $delta_index,
        $field_property_name
      );
      $field_values[] = reset($values);
    }
    return $field_values;
  }

}
