<?php

namespace Drupal\external_entities\Plugin\ExternalEntities\StorageClient;

use Drupal\Component\Utility\Random;
use Drupal\Core\Form\FormStateInterface;

/**
 * External entities storage client based on a JSON:API.
 *
 * @StorageClient(
 *   id = "jsonapi",
 *   label = @Translation("JSON:API"),
 *   description = @Translation("Retrieves external entities from a (Drupal) JSON:API source.")
 * )
 */
class JsonApi extends Rest {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form = parent::buildConfigurationForm($form, $form_state);

    // Adjust default REST form.
    $form['endpoint']['#title'] = $this->t('JSON:API URL');
    $form['endpoint']['#description'] = $this->t(
      'Should be something similar to "<code>http://www.server.org/jsonapi/content_type/bundle_name/</code>". Ex.: <code>http://www.server.org/jsonapi/node/page/</code>.'
    );
    $form['endpoint']['#attributes']['#placeholder'] = $this->t("ex.: http://www.someserver.net/jsonapi/node/page/");

    $form['endpoint_options']['cache'] = [
      '#type' => 'hidden',
      '#value' => TRUE,
    ];

    $form['endpoint_options']['single'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    $form['endpoint_options']['count'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Estimated total number of entities'),
      '#required' => FALSE,
      '#description' => $this->t('Since JSON:API does not provide a mean to get the total number of available entities, you should provide a static value here that correspond to the estimated total number of available entities. This number could be higher than reality as it will only be used for paging computation.'),
      '#default_value' => $this->configuration['endpoint_options']['count'] ?? '',
    ];

    $form['endpoint_options']['count_mode'] = [
      '#type' => 'hidden',
      '#value' => 'entities',
    ];

    $form['endpoint_options']['limit_qcount']['#title'] = $this->t('Maximum number of query per second');
    $form['endpoint_options']['limit_qcount']['#description'] = $this->t('JSON:API does not handle query rate but it is still possible to limit query rate from this client side.');

    $form['endpoint_options']['limit_qtime'] = [
      '#type' => 'hidden',
      '#value' => '1',
    ];

    $form['response_format'] = [
      '#type' => 'hidden',
      '#value' => 'json',
    ];

    unset($form['data_path']['#type']);

    $form['data_path']['list'] = [
      '#type' => 'hidden',
      '#value' => '$.data.*',
    ];

    $form['data_path']['single'] = [
      '#type' => 'hidden',
      '#value' => '$[data,included]',
    ];

    $form['data_path']['keyed_by_id'] = [
      '#type' => 'hidden',
      '#value' => FALSE,
    ];

    $form['data_path']['count'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    unset($form['pager']['#type']);

    // Drupal uses a 50 elements per bage basis so we go for 1 query per page.
    // 50 is the JSON:API default limit anyway.
    $form['pager']['default_limit'] = [
      '#type' => 'hidden',
      '#value' => static::DEFAULT_PAGE_LENGTH,
    ];

    $form['pager']['page_parameter'] = [
      '#type' => 'hidden',
      '#value' => 'page[offset]',
    ];

    $form['pager']['page_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'startitem',
    ];

    $form['pager']['page_size_parameter'] = [
      '#type' => 'hidden',
      '#value' => 'page[limit]',
    ];

    $form['pager']['page_size_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'pagesize',
    ];

    $form['pager']['page_start_one'] = [
      '#type' => 'hidden',
      '#value' => FALSE,
    ];

    unset($form['api_key']['#type']);

    $form['api_key']['type'] = [
      '#type' => 'hidden',
      '#value' => 'none',
    ];

    $form['api_key']['key'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    $form['api_key']['header_name'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    unset($form['http']['#type']);

    $form['http']['headers'] = [
      '#type' => 'hidden',
      '#value' => 'Accept: application/vnd.api+json',
    ];

    $form['parameters']['#description'] = $this->t('You may specify here additional URL query parameters that will be used to filter entities or include extra data for instance. You can find available parameters in the <a href="https://www.drupal.org/docs/core-modules-and-themes/core-modules/jsonapi-module">JSON:API documentation</a>.');

    $form['parameters']['single']['#title'] = $this->t('Additional parameters');
    $form['parameters']['single']['#attributes']['placeholder'] = $this->t("ex.:\nfilter[field_name]|value\nfilter[field_other]|value\ninclude=field_comments.uid");
    $form['parameters']['list_param_mode'] = [
      '#type' => 'hidden',
      '#value' => 'query',
    ];
    $form['parameters']['list'] = [
      '#type' => 'hidden',
      '#value' => $this->getParametersFormDefaultValue('list'),
    ];
    $form['parameters']['single_param_mode'] = [
      '#type' => 'hidden',
      '#value' => 'query',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $parameters = $form_state->getValue('parameters');
    if (preg_match('#(?:^|\n)include\|([^\n]+)#', $parameters['single'])) {
      // We have "include". Remove any from the "list" and disable static cache.
      $parameters['list'] = preg_replace('#(?<=^|\n)(?:include)\|[^\n]*\n?#', '', $parameters['single']);
      $form_state->setValue(['endpoint_options', 'cache'], FALSE);
    }
    else {
      // Same parameters for both single and list if no "include".
      $parameters['list'] = $parameters['single'];
    }
    $form_state->setValue('parameters', $parameters);
    parent::submitConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   *
   * @see https://www.drupal.org/docs/core-modules-and-themes/core-modules/jsonapi-module/filtering
   */
  public function transliterateDrupalFilters(array $parameters) :array {
    // If no external entity is available, no way to map fields.
    if (empty($this->externalEntityType)) {
      return [
        'source' => [],
        'drupal' => $parameters,
      ];
    }
    $field_mapper = $this->externalEntityType->getFieldMapper();

    $source_filters = [];
    $drupal_filters = [];
    foreach ($parameters as $parameter) {
      if (!isset($parameter['field']) || !isset($parameter['value'])) {
        $drupal_filters[] = $parameter;
        continue;
      }
      $source_field = $field_mapper->getMappedFieldName($parameter['field']);
      if (!isset($source_field)) {
        $drupal_filters[] = $parameter;
        continue;
      }

      // We are probably dealing with a filter condition.
      // The structure the filter needs to be the following:
      // @example ?filter[recent][condition][path]=changed&filter[recent][condition][operator]=>&filter[recent][condition][value]=1712670000.
      $randomizer = new Random();
      $random_name = $randomizer->name();
      $mapped_parameter['filter[' . $random_name . '][condition][path]'] = $source_field;
      $mapped_parameter['filter[' . $random_name . '][condition][operator]'] = $parameter['operator'] ?? '=';
      $mapped_parameter['filter[' . $random_name . '][condition][value]'] = $parameter['value'];
      $mapped_parameters[] = $mapped_parameter;
    }

    return ['source' => $source_filters, 'drupal' => $drupal_filters];
  }

  /**
   * {@inheritdoc}
   */
  public function load(string|int $id) :array|null {
    // Check static cache.
    if (!empty(static::$cachedData[$this->configuration['endpoint']][$id])) {
      return static::$cachedData[$this->configuration['endpoint']][$id];
    }

    $result = parent::load($id);

    // Check for "included".
    if (!empty($result[1])) {
      // JSONPath returned an array of object so we got "extra" from "included".
      $included = array_slice($result, 1);
      $result = $result[0];
      $result['included'] = $included;
    }
    static::$cachedData[$this->configuration['endpoint']][$id] = $result;

    return $result;
  }

}
