<?php

namespace Drupal\external_entities\Plugin\ExternalEntities\StorageClient;

use Drupal\Component\Serialization\Exception\InvalidDataTypeException;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\external_entities\ExternalEntityInterface;
use Drupal\external_entities\Plugin\PluginFormTrait;
use Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface;
use Drupal\external_entities\StorageClient\StorageClientBase;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\GuzzleException;
use JsonPath\InvalidJsonException;
use JsonPath\InvalidJsonPathException;
use JsonPath\JsonObject;
use Psr\Http\Message\ResponseInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Yaml\Yaml;

/**
 * External entities storage client based on a REST API.
 *
 * @StorageClient(
 *   id = "rest",
 *   label = @Translation("REST"),
 *   description = @Translation("Retrieves external entities from a REST API.")
 * )
 */
class Rest extends StorageClientBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * Default page length for endpoint queries if not specified.
   *
   * (default set to 50 as it's Drupal entity list default size)
   * See \Drupal\Core\Entity\EntityListBuilder::$limit.
   *
   * @var int
   */
  public const DEFAULT_PAGE_LENGTH = 50;

  /**
   * The response decoder factory.
   *
   * @var \Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface
   */
  protected $responseDecoderFactory;

  /**
   * The HTTP client to fetch the files with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerChannelFactory;

  /**
   * Data static cache.
   *
   * First level key are endpoints, second level are item numbers.
   *
   * @var array
   */
  protected static $cachedData;

  /**
   * Database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * Constructs a Rest object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager service.
   * @param \Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface $response_decoder_factory
   *   The response decoder factory service.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   A Guzzle client object.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger channel factory.
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    EntityFieldManagerInterface $entity_field_manager,
    ResponseDecoderFactoryInterface $response_decoder_factory,
    ClientInterface $http_client,
    LoggerChannelFactoryInterface $logger_factory,
    Connection $connection
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $string_translation, $entity_field_manager);
    $this->httpClient = $http_client;
    $this->loggerChannelFactory = $logger_factory;
    $this->responseDecoderFactory = $response_decoder_factory;
    $this->connection = $connection;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('entity_field.manager'),
      $container->get('external_entities.response_decoder_factory'),
      $container->get('http_client'),
      $container->get('logger.factory'),
      $container->get('database')
    );
  }

  /**
   * Returns the response decoder factory.
   *
   * @return \Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface
   *   The response decoder factory.
   */
  public function getResponseDecoderFactory() {
    return $this->responseDecoderFactory;
  }

  /**
   * Sets the response decoder factory.
   *
   * @param \Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface $response_decoder_factory
   *   A response decoder factory.
   *
   * @return $this
   */
  public function setResponseDecoderFactory(ResponseDecoderFactoryInterface $response_decoder_factory) {
    $this->responseDecoderFactory = $response_decoder_factory;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'endpoint' => NULL,
      'endpoint_options' => [
        'single' => NULL,
        'count' => NULL,
        'count_mode' => NULL,
        'cache' => NULL,
        'limit_qcount' => NULL,
        'limit_qtime' => NULL,
      ],
      'response_format' => NULL,
      'data_path' => [
        'list' => NULL,
        'single' => NULL,
        'keyed_by_id' => NULL,
        'count' => NULL,
      ],
      'pager' => [
        'default_limit' => NULL,
        'page_parameter' => NULL,
        'page_parameter_type' => NULL,
        'page_start_one' => NULL,
        'page_size_parameter' => NULL,
        'page_size_parameter_type' => NULL,
      ],
      'api_key' => [
        'type' => NULL,
        'header_name' => NULL,
        'key' => NULL,
      ],
      'http' => [
        'headers' => NULL,
      ],
      'parameters' => [
        'list' => NULL,
        'list_param_mode' => NULL,
        'single' => NULL,
        'single_param_mode' => NULL,
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    // Here we use a generated identifier to be able to locate the form element
    // for the states API because we can not rely on its form name since it is
    // embeded in subforms and may aslo be duplicated (when multiple REST
    // sources are mixed).
    $rest_id = uniqid('rest', TRUE);

    $form['endpoint'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Endpoint URL'),
      '#required' => TRUE,
      '#description' => $this->t('Main endpoint URL used to list all available entities.'),
      '#default_value' => $this->configuration['endpoint'] ?? '',
      '#attributes' => [
        'placeholder' => $this->t('ex.: https://www.server.org/service/v1/entities/'),
      ],
    ];

    $got_ep_options =
      !empty($this->configuration['endpoint_options']['cache'])
      || !empty($this->configuration['endpoint_options']['single'])
      || !empty($this->configuration['endpoint_options']['count']);
    $form['endpoint_options'] = [
      '#type' => 'details',
      '#title' => $this->t('Endpoint options'),
      '#open' => $got_ep_options,
    ];

    $form['endpoint_options']['cache'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('The main endpoint provides full entities in a list.'),
      '#description' => $this->t('If the main endpoint URL used to list entities provides full entities (same fields as the ones provided by the endpoint URL used to load a single entity), checking this box will improve performances (caching listed entities).'),
      '#default_value' => $this->configuration['endpoint_options']['cache'] ?? FALSE,
    ];

    $form['endpoint_options']['single'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Endpoint URL for a single entity'),
      '#required' => FALSE,
      '#description' => $this->t('Endpoint URL used to access a single entity if different from the main endpoint given above. This URL should contain a "{id}" placeholder that will be replaced by the entity identifier. If left blank, a single entity\'s URL will be built on the main endpoint URL with the "/entity_id" appended. Note: entity identifiers are URL-encoded.'),
      '#default_value' => $this->configuration['endpoint_options']['single'] ?? '',
      '#attributes' => [
        'placeholder' => $this->t('ex.: https://www.server.org/service/v1/entities/{id}/details'),
      ],
    ];

    $form['endpoint_options']['count'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Endpoint URL to get the total number of available entities or pages'),
      '#required' => FALSE,
      '#description' => $this->t('Endpoint URL used to return the total number of entities (or pages) that can be fetched through the main endpoint if the later does not return a count. The endpoint should return either a numeric value, an array containing a numeric value, a list containing a number of items equal to the total number of entities or a complex structure that will be processed by a JSONPath leading to the total number of available entities specified below. It is also possible to directly specify a static total number (as integer) of entities instead of a URL.'),
      '#default_value' => $this->configuration['endpoint_options']['count'] ?? '',
      '#attributes' => [
        'placeholder' => $this->t('ex.: "https://www.server.org/service/v1/count/entities" or "25000"'),
      ],
    ];

    $form['endpoint_options']['count_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Type of total number of items returned'),
      '#options' => [
        'entities' => $this->t('Number of entities'),
        'pages' => $this->t('Number of pages'),
      ],
      '#default_value' => $this->configuration['endpoint_options']['count_mode'] ?? 'entities',
    ];

    // Handles query limitations.
    $form['endpoint_options']['limit_qcount'] = [
      '#type' => 'number',
      '#title' => $this->t('Maximum number of query per amount of time given below'),
      '#description' => $this->t('Some endpoints restrict the number of queries for a certain amount of time. While this setting will not guarantee the limitation will be respected (since multiple users could access the endpoint from this site in parallel sessions), it could help limiting abuses. Leave this field empty or set to 0 if there are no known limitations.'),
      '#default_value' => $this->configuration['endpoint_options']['limit_qcount'] ?? 0,
      '#min' => 0,
    ];

    $form['endpoint_options']['limit_qtime'] = [
      '#type' => 'number',
      '#title' => $this->t('Amount of time given to limit queries (in seconds)'),
      '#description' => $this->t('If your endpoint provides a time value in minutes our even hours, convert it into seconds.'),
      '#default_value' => $this->configuration['endpoint_options']['limit_qtime'] ?? 0,
      '#min' => 0,
      '#field_suffix' => $this->t('seconds'),
    ];

    $formats = $this->responseDecoderFactory->supportedFormats();
    $form['response_format'] = [
      '#type' => 'select',
      '#title' => $this->t('Endpoint data encoding format'),
      '#options' => array_combine($formats, $formats),
      '#required' => TRUE,
      '#default_value' => $this->configuration['response_format'] ?? '',
    ];

    $form['data_path'] = [
      '#type' => 'details',
      '#title' => $this->t('Data path settings'),
      '#description' => $this->t('Sometimes REST endpoints return their entities with metadata. Entity lists and entities may therefore not be found at the first level of the response. The following settings help to handle thoses cases. They are based on <a href="https://goessner.net/articles/JsonPath/">JSONPath</a> (<a href="https://github.com/Galbar/JsonPath-PHP">PHP implementation</a>) to locate data items.'),
      '#open' => (bool) $this->configuration['data_path']['list'],
    ];

    $form['data_path']['list'] = [
      '#type' => 'textfield',
      '#title' => $this->t('JSONPath leading to the list of entities'),
      '#description' => $this->t('Use this field when entities are not returned in a simple list but rather in a more complex structure. Ex.: if the REST endpoint returns entities wrapped in a structure like <code>{..., "data": [{object1},{object2},...]}</code>, use the JSON path <code>"$.data.*"</code>.'),
      '#default_value' => $this->configuration['data_path']['list'] ?? '',
      '#attributes' => [
        'placeholder' => $this->t('ex.: $.data.*'),
      ],
    ];

    $form['data_path']['single'] = [
      '#type' => 'textfield',
      '#title' => $this->t('JSONPath leading to the content of a single entity'),
      '#description' => $this->t('Use this field when entities are not returned directly but rather in a sub-structure. Ex.: if the REST endpoint returns the entity wrapped in a structure like <code>{..., "item": {object1}}</code>, use the JSON path <code>"$.item"</code>.'),
      '#default_value' => $this->configuration['data_path']['single'] ?? '',
      '#attributes' => [
        'placeholder' => $this->t('ex.: $.data.0'),
      ],
    ];

    $form['data_path']['keyed_by_id'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Entities are keyed by their ID in a list.'),
      '#description' => $this->t('Check this box if the endpoints returns a list of entities keyed by their identifiers (note: not RESTful compliant, will not work for data editing). If the returned entities do not have an "id" field, an "id" field will be set with the corresponding key value.'),
      '#default_value' => $this->configuration['data_path']['keyed_by_id'] ?? FALSE,
    ];

    $form['data_path']['count'] = [
      '#type' => 'textfield',
      '#title' => $this->t('JSONPath leading to the total number of available entities'),
      '#description' => $this->t('This JSONPath will be used against the result of the count query provided above or against the list endpoint otherwise.'),
      '#default_value' => $this->configuration['data_path']['count'] ?? '',
      '#attributes' => [
        'placeholder' => $this->t('ex.: $.metadata.total'),
      ],
    ];

    $form['pager'] = [
      '#type' => 'details',
      '#title' => $this->t('Pager settings'),
      '#open' => TRUE,
    ];

    $form['pager']['default_limit'] = [
      '#type' => 'number',
      '#title' => $this->t('Default number of items per page used by the endpoint or maximum number of entities supported in a single query.'),
      '#description' => $this->t('This number will be used to compute paging on the Drupal side. It has nothing to deal with the number of entities displayed on Drupal entity list pages.'),
      '#default_value' => $this->configuration['pager']['default_limit'] ?? '',
      '#min' => 0,
    ];

    $form['pager']['page_parameter'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Page parameter'),
      '#default_value' => $this->configuration['pager']['page_parameter'] ?? '',
    ];

    // @todo Maybe support the "previous page/next page/last page" paging type?
    //   This is usually handled by links either in the response metadata or in
    //   the HTTP header ('link: <https://some.url/...>; rel="next", ...').
    //   Sometimes, the metadata does not contain a full link but rather a token
    //   to use in the next query with some parameter name.
    //   So if we add "Page parameter type" called "previous/next/last paging":
    //   - we should add a new 'data_path' setting for 'next page data' that
    //     would allow to choose between a JSONPath or an HTTP header link with
    //     the relation name of the link that holds the next page
    //   - if 'page_parameter' is not set and the token value begins with http,
    //     we should treate the token as the URL to use to get the next set of
    //     results
    //   - we should handle 'page_parameter' if it is set, as the parameter name
    //     to append to the query with the token obtained from a
    //     'next page data' setting
    //   All those settings would be used in ::getPagingQueryParameters() to
    //   load the apprioriate pages but it will need caching as we would loose
    //   the last token from one Drupal page query to another.
    $form['pager']['page_parameter_type'] = [
      '#type' => 'radios',
      '#title' => $this->t('Page parameter type'),
      '#options' => [
        'pagenum' => $this->t('Page number'),
        'startitem' => $this->t('Starting item'),
      ],
      '#description' => $this->t('Use "Page number" when the pager uses page numbers to determine the item to start at, use "Starting item" when the pager uses the item number to start at.'),
      '#default_value' => $this->configuration['pager']['page_parameter_type'] ?? '',
    ];

    $form['pager']['page_size_parameter'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Page size parameter'),
      '#default_value' => $this->configuration['pager']['page_size_parameter'] ?? '',
    ];

    $form['pager']['page_size_parameter_type'] = [
      '#type' => 'radios',
      '#title' => $this->t('Page size parameter type'),
      '#options' => [
        'pagesize' => $this->t('Number of items per page'),
        'enditem' => $this->t('Ending item'),
      ],
      '#description' => $this->t('Use "Number of items per pager" when the pager uses this parameter to determine the amount of items on each page, use "Ending item when the pager uses this parameter to determine the number of the last item on the page.'),
      '#default_value' => $this->configuration['pager']['page_size_parameter_type'] ?? '',
    ];

    $form['pager']['page_start_one'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Page numbers and entity numbers start from one.'),
      '#description' => $this->t('Check this box if the first page number or entity number is 1, uncheck it if they start from 0.'),
      '#default_value' => $this->configuration['pager']['page_start_one'] ?? FALSE,
    ];

    $form['api_key'] = [
      '#type' => 'details',
      '#title' => $this->t('Authentication'),
      '#open' => (!empty($this->configuration['api_key']['type'])
          && ('none' != $this->configuration['api_key']['type'])),
    ];

    $form['api_key']['type'] = [
      '#type' => 'radios',
      '#title' => $this->t('Authentication type'),
      '#default_value' => $this->configuration['api_key']['type'] ?? 'none',
      '#options' => [
        'none'   => $this->t('No authentication needed'),
        'bearer' => $this->t('Bearer authentication (also called token authentication, often used by OAuth)'),
        'custom' => $this->t('Custom header'),
        'query' => $this->t('Using a URL query parameter'),
        // @todo Add support for authentication plugins and add their options:
        // "'plugin_id' => $this->t('Using authentication <plugin_name>'),".
      ],
      '#attributes' => [
        'data-xnttrest-selector' => 'atype' . $rest_id,
      ],
    ];

    $form['api_key']['header_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Header name/key name'),
      '#description' => $this->t('The HTTP header name or the URL query parameter name for the API key.'),
      '#default_value' => $this->configuration['api_key']['header_name'] ?? '',
      '#attributes' => [
        'placeholder' => $this->t('ex.: "WWW-Authenticate" or "apikey"'),
      ],
      '#states' => [
        'visible' => [
          'input[data-xnttrest-selector="atype' . $rest_id . '"]' => [
            ['value' => 'custom'],
            'or',
            ['value' => 'query'],
          ],
        ],
      ],
    ];

    $form['api_key']['key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API key'),
      '#description' => $this->t('The API key needed to communicate with the entered endpoint.'),
      '#default_value' => $this->configuration['api_key']['key'] ?? '',
      '#states' => [
        'invisible' => [
          'input[data-xnttrest-selector="atype' . $rest_id . '"]' => ['value' => 'none'],
        ],
      ],
    ];
    // Remove 'Bearer ' part automatically added by form validation for 'bearer'
    // mode.
    if (!empty($this->configuration['api_key']['type'])
        && ('bearer' == $this->configuration['api_key']['type'])
    ) {
      $form['api_key']['key']['#default_value'] = str_replace(
        'Bearer ',
        '',
        $form['api_key']['key']['#default_value']
      );
    }

    $form['http'] = [
      '#type' => 'details',
      '#title' => $this->t('HTTP protocol'),
      '#open' => FALSE,
    ];
    $form['http']['headers'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Additional HTTP headers'),
      '#description' => $this->t('You may specify hear additional HTTP headers to include in the queries.'),
      '#default_value' => $this->configuration['http']['headers'] ?? '',
    ];

    $list_parameters = $this->getParametersFormDefaultValue('list');
    $single_parameters = $this->getParametersFormDefaultValue('single');
    $form['parameters'] = [
      '#type' => 'details',
      '#title' => $this->t('Additional parameters'),
      '#description' => $this->t('You may specify here additional URL query parameters that will be added when listing entities or fetching a single entity. In the case of a single entity query, the placeholder "{id}" will be replaced by the entity identifier in values. Please note that you can also specify a "query" API key using the authentication section above (prefered way) which will automatically add the key to all types of URL query.'),
      '#open' => (!empty($list_parameters) || !empty($single_parameters)),
    ];

    $form['parameters']['list'] = [
      '#type' => 'textarea',
      '#title' => $this->t('List parameters'),
      '#description' => $this->t('Enter the parameters to add to the endpoint URL when loading the list of entities. One per line in the format "parameter_name|parameter_value".'),
      '#default_value' => $list_parameters,
    ];
    $form['parameters']['list_param_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Add list parameters to'),
      '#default_value' => $this->configuration['parameters']['list_param_mode'] ?? 'query',
      '#options' => [
        'query' => $this->t('query string (default)'),
        'body' => $this->t('body (using selected endpoint response format)'),
        'body_query' => $this->t('body for list and query string for count'),
      ],
      '#description' => $this->t('Sometimes, additional parameters need to be sent in the request body when searching/filtering entities. When set to "body", list and count queries will use "POST" HTTP method instead of "GET". It is possible to mix both by specifying query parameters in the endpoint URL and specify body parameters here.'),
      '#attributes' => [
        'data-xnttrest-selector' => 'lpmode' . $rest_id,
      ],
    ];

    $form['parameters']['single'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Single entity parameters'),
      '#description' => $this->t('Enter the parameters to add to the endpoint URL when loading a single entity. One per line in the format "parameter_name|parameter_value". Placeholder "{id}" in parameter values will be replaced by the entity identifier.'),
      '#default_value' => $single_parameters,
    ];

    $form['parameters']['single_param_mode'] = [
      '#type' => 'radios',
      '#title' => $this->t('Add single entity parameters to'),
      '#default_value' => $this->configuration['parameters']['single_param_mode'] ?? 'query',
      '#options' => [
        'query' => $this->t('query string (default)'),
        'body' => $this->t('body (using selected endpoint response format)'),
      ],
      '#description' => $this->t('Sometimes, additional parameters need to be sent in the request body when saving an entity (POST, PUT). It is possible to mix both by specifying query parameters in the endpoint URL and specify body parameters here.'),
      '#attributes' => [
        'data-xnttrest-selector' => 'spmode' . $rest_id,
      ],
    ];

    return $form;
  }

  /**
   * Helper function to convert a parameter collection to a string.
   *
   * @param string $type
   *   The type of parameters (eg. 'list' or 'single').
   *
   * @return string
   *   A string to be used as default value, an empty string if no parameters.
   */
  protected function getParametersFormDefaultValue($type) :string {
    $default_value = '';

    if (!empty($this->configuration['parameters'][$type])) {
      $lines = [];
      foreach ($this->configuration['parameters'][$type] as $key => $value) {
        $array = array_unique([$key, $value]);
        $lines[] = implode('|', $array);
      }
      $default_value = implode("\n", $lines);
    }

    return $default_value;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    // We need to get parameters value to check single endpoint requirements.
    $parameters = $form_state->getValue('parameters') ?? [];

    $ep_options = $form_state->getValue('endpoint_options') ?? [];
    // Check {id} placeholder.
    if (!empty($ep_options['single'])
        && !preg_match('~\{id\}~', $ep_options['single'])
        && (empty($parameters['single'])
          || !preg_match('~\{id\}~', $parameters['single']))
    ) {
      // Warn the single enpoint URL may not be correct because of missing {id}
      // placeholder both in URL and parameters.
      \Drupal::messenger()->addWarning(
        'The endpoint for a single entity does not contain a "{id}" placeholder ('
        . $ep_options['single']
        . ').'
      );
    }

    if (!is_numeric($ep_options['limit_qcount'])) {
      $ep_options['limit_qcount'] = 0;
    }

    if (!is_numeric($ep_options['limit_qtime'])) {
      $ep_options['limit_qtime'] = 0;
    }

    // Rearrange parameters.
    foreach (['list', 'single'] as $type) {
      // Make sure the parameters have not been already processed by a child
      // module.
      $value = $parameters[$type];
      if (is_string($value)) {
        $lines = explode("\n", $value);
        $lines = array_map('trim', $lines);
        $lines = array_filter($lines, 'strlen');
        $parameters[$type] = [];
        foreach ($lines as $line) {
          $exploded = explode('|', $line);
          $value = !empty($exploded[1]) ? $exploded[1] : $exploded[0];
          $parameters[$type][$exploded[0]] = $value;
        }
      }
    }

    $api_key = $form_state->getValue('api_key');

    // Bearer: prefix key if needed and set header name.
    if (!empty($api_key['type'])
        && ('bearer' == $api_key['type'])
        && !empty($api_key['key'])
        && (@substr_compare($api_key['key'], 'Bearer', 0, 6) !== 0)
    ) {
      $api_key['key'] = 'Bearer ' . $api_key['key'];
      $api_key['header_name'] = 'Authorization';
    }

    // Check authentication settings.
    if (!empty($api_key['type'])
        && ('none' != $api_key['type'])
    ) {
      if (empty($api_key['key']) || empty($api_key['header_name'])) {
        // Warn for incorrect setting.
        \Drupal::messenger()->addWarning('Authentication mode set but no authentication key properly set. Authentication disabled.');
        $api_key['type'] = 'none';
      }
      // Check for secure protocol.
      $endpoint = $form_state->getValue('endpoint');
      if (!preg_match('#^https://#i', $endpoint)
        || (!empty($ep_options['single'])
            && !preg_match('#^https://#i', $ep_options['single']))
        || (!empty($ep_options['count'])
            && !preg_match('#^https://#i', $ep_options['count']))
      ) {
        // Warn for security issues.
        \Drupal::messenger()->addWarning('Authentication mode set but endpoint URL(s) are not using secure protocol (https). The API key may be intercepted by others. Consider switching to https unless the key is not secret.');
      }
    }
    // Save endpoint options settings.
    $form_state->setValue('endpoint_options', $ep_options);

    // Save authentication settings.
    $form_state->setValue('api_key', $api_key);

    // Save parameters rearrangement.
    $form_state->setValue('parameters', $parameters);

    $this->setConfiguration($form_state->getValues());
  }

  /**
   * Returns the URL of a single entity.
   *
   * @param string|int $id
   *   Entity identifier.
   *
   * @return string
   *   The endpoint URL to use to fetch the entity. Any '{id}' placeholders are
   *   replaced by the given identifier. If the settings provide a single entity
   *   URL without any '{id}' placeholder, that URL will be returned as is. If
   *   no single entity URL was provided, the list endpoint will be used with an
   *   appended '/' + the provided identifier.
   */
  public function getSingleEntityUrl(string|int $id) :string {
    // Get single entity URL either from a given specific URL or the general
    // endpoint URL.
    $endpoint =
      $this->configuration['endpoint_options']['single']
      ?: (rtrim($this->configuration['endpoint'], '/') . '/' . urlencode($id));

    // Replace any {id} placeholder with the given identifier (URL-encoded).
    $endpoint = str_replace('{id}', urlencode($id), $endpoint);

    // Strip query string part as it will be managed by
    // getListQueryParameters().
    $endpoint = strstr($endpoint . '?', '?', TRUE);

    return $endpoint;
  }

  /**
   * {@inheritdoc}
   */
  public function delete(ExternalEntityInterface $entity) {

    $endpoint = $this->getSingleEntityUrl($entity->id());
    $method = $this->getRequestMethod('delete', $endpoint);
    try {
      $parameters = NestedArray::mergeDeep(
        $this->getSingleBodyParameters($entity->id()),
        [
          'headers' => $this->getHttpHeaders(),
          'query' => $this->getSingleQueryParameters($entity->id()),
        ]
      );
      $this->ensureQueryLimits();
      $this->httpClient->request(
        $method,
        $endpoint,
        $parameters
      );
    }
    catch (GuzzleException $exception) {
      $logger = $this->loggerChannelFactory->get('external_entities');
      $logger->error('Error in Rest::delete. Message: ' . $exception->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(array $ids = NULL) :array {
    $data = [];

    if (!empty($ids) && is_array($ids)) {
      foreach ($ids as $id) {
        $endpoint_data = $this->load($id);
        if (empty($endpoint_data)) {
          continue;
        }
        if (!empty($endpoint_data[0])) {
          $endpoint_data = reset($endpoint_data);
        }
        $data[$id] = $endpoint_data;
      }
    }

    return $data;
  }

  /**
   * Loads one entity.
   *
   * @param mixed $id
   *   The ID of the entity to load.
   *
   * @return array
   *   A raw data array, empty array if no data returned.
   */
  public function load(string|int $id) :array|null {
    // Check static cache.
    if (!empty(static::$cachedData[$this->configuration['endpoint']][$id])) {
      return static::$cachedData[$this->configuration['endpoint']][$id];
    }
    $endpoint = $this->getSingleEntityUrl($id);
    $method = $this->getRequestMethod('read', $endpoint);
    $parameters = NestedArray::mergeDeep(
      $this->getSingleBodyParameters($id),
      [
        'headers' => $this->getHttpHeaders(),
        'query' => $this->getSingleQueryParameters($id),
      ]
    );
    try {
      $this->ensureQueryLimits();
      $response = $this->httpClient->request(
        $method,
        $endpoint,
        $parameters
      );
    }
    catch (GuzzleException $exception) {
      $logger = $this->loggerChannelFactory->get('external_entities');
      $logger->error('Error in Rest::load. Message: ' . $exception->getMessage());
      return [];
    }

    $body = $this->getResponseBody(
      $response,
      'read',
      $method,
      $endpoint,
      $parameters
    );

    $result = [];
    try {
      $result = $this
        ->getResponseDecoderFactory()
        ->getDecoder($this->configuration['response_format'])
        ->decode($body);
    }
    catch (InvalidDataTypeException $exception) {
      $logger = $this->loggerChannelFactory->get('external_entities');
      $logger->error('Error in decoding result. Message: ' . $exception->getMessage());
    }

    // Check if we must use a JSON Path to get data.
    if (!empty($this->configuration['data_path']['single'])) {
      try {
        $json_object = new JsonObject($result, TRUE);
        $result = $json_object->get($this->configuration['data_path']['single']);
        if (FALSE === $result) {
          $result = [];
        }
      }
      catch (InvalidJsonException | InvalidJsonPathException $e) {
        $logger = $this->loggerChannelFactory->get('external_entities');
        $logger->error(
          'Failed to parse JSON Path for single object ('
          . $this->configuration['data_path']['single']
          . '): '
          . $e
        );
      }

      // Check if entities are keyed by their ids.
      if (!empty($this->configuration['data_path']['keyed_by_id'])) {
        if (!empty($result[$id])) {
          $result = $result[$id];
          $result['id'] ??= $id;
        }
        else {
          $result = [];
        }
      }
    }

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function save(ExternalEntityInterface $entity) :int {
    $result = [];
    try {
      if ($entity->id()) {
        $endpoint = $this->getSingleEntityUrl($entity->id());
        $method = $this->getRequestMethod('update', $endpoint);
        $parameters = NestedArray::mergeDeep(
          $this->getSingleBodyParameters($entity->id(), $entity->toRawData()),
          [
            'headers' => $this->getHttpHeaders(),
            'query' => $this->getSingleQueryParameters($entity->id()),
          ]
        );
        $this->ensureQueryLimits();
        $this->httpClient->request(
          $method,
          $endpoint,
          $parameters
        );
        $result = SAVED_UPDATED;
      }
      else {
        $endpoint = $this->getSingleEntityUrl('');
        $method = $this->getRequestMethod('create', $endpoint);
        $parameters = NestedArray::mergeDeep(
          $this->getSingleBodyParameters($entity->id(), $entity->toRawData()),
          [
            'headers' => $this->getHttpHeaders(),
            'query' => $this->getSingleQueryParameters($entity->id()),
          ]
        );
        $this->ensureQueryLimits();
        $this->httpClient->request(
          $method,
          $endpoint,
          $parameters
        );
        $result = SAVED_NEW;
      }
    }
    catch (GuzzleException $exception) {
      $logger = $this->loggerChannelFactory->get('external_entities');
      $logger->error('Error in Rest::save. Message: ' . $exception->getMessage());
      return [];
    }

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function querySource(
    array $parameters = [],
    array $sorts = [],
    ?int $start = NULL,
    ?int $length = NULL
  ) :array {
    $all_results = [];
    $endpoint = $this->configuration['endpoint'];
    $method = $this->getRequestMethod('query', $endpoint);
    foreach ($this->getPagingQueryParameters($start, $length) as $i => $paging) {
      try {
        $req_parameters = NestedArray::mergeDeep(
          $this->getListBodyParameters($parameters),
          [
            'headers' => $this->getHttpHeaders(),
            'query' => $paging['parameters']
            + $this->getListQueryParameters($parameters),
          ]
        );
        // Manage delays between queries.
        $this->ensureQueryLimits();
        $response = $this->httpClient->request(
          $method,
          $endpoint,
          $req_parameters
        );
      }
      catch (GuzzleException $exception) {
        $logger = $this->loggerChannelFactory->get('external_entities');
        $logger->error('Error in Rest::query. Message: ' . $exception->getMessage());
        return [];
      }

      $body = $this->getResponseBody(
        $response,
        'list',
        $method,
        $endpoint,
        $req_parameters
      );

      try {
        $results = $this
          ->getResponseDecoderFactory()
          ->getDecoder($this->configuration['response_format'])
          ->decode($body);
      }
      catch (InvalidDataTypeException $exception) {
        $logger = $this->loggerChannelFactory->get('external_entities');
        $logger->error('Error in decoding result. Message: ' . $exception->getMessage());
        return [];
      }

      // Check if we must use a JSON Path to get data.
      if (!empty($this->configuration['data_path']['list'])) {
        try {
          $json_object = new JsonObject($results, TRUE);
          $results = $json_object->get($this->configuration['data_path']['list']);
          if (FALSE === $results) {
            $results = [];
          }
        }
        catch (InvalidJsonException | InvalidJsonPathException $e) {
          $logger = $this->loggerChannelFactory->get('external_entities');
          $logger->error(
            'Failed to parse JSON Path for listing ('
            . $this->configuration['data_path']['list']
            . '): '
            . $e
          );
        }
      }

      // @todo Fix for DYNAMIC COUNT: For endpoints with no count, check if we
      // reached the end. First, make sure there are no filtering parameters as
      // we don't support dynamic count when filtering (too many cases).
      // If the page contains less results than expected, it means we reached
      // the end, so update the cache (maybe also the config?)
      // 'external_entities_<entity_type>_rest_count' to the correct count.
      // If we got enough results, check if we got more than requested on the
      // last page or if it is full and if we reached current count in the
      // cache. If so, increase the cache by one page size.
      $all_results = array_merge(
        $all_results,
        array_slice($results, $paging['start'], $paging['length'])
      );
    }

    // Check if results are just ids.
    if (!empty($all_results) && !is_array($all_results[0])) {
      $all_results = array_map(
        function ($item) {
          // Get the source identifier field name.
          $source_id_field = $this->getSourceIdFieldName() ?? 'id';
          return [
            $source_id_field => $item,
          ];
        },
        $all_results
      );
    }

    if (!empty($all_results) && !empty($this->configuration['endpoint_options']['cache'])) {
      $field_mapper = $this->externalEntityType->getFieldMapper();
      static::$cachedData[$this->configuration['endpoint']] ??= [];
      foreach ($all_results as $raw_data) {
        if (!empty($raw_data)) {
          $id = $field_mapper->extractIdFromRawData($raw_data);
          if (isset($id)) {
            // No warning if $all_results[$i] is not set.
            static::$cachedData[$this->configuration['endpoint']][$id] = $raw_data;
          }
        }
      }
    }

    return $all_results;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuerySource(array $parameters = []) :int {
    if (!empty($this->configuration['endpoint_options']['count'])
        && is_numeric($this->configuration['endpoint_options']['count'])
    ) {
      // Static count.
      // @todo Fix for DYNAMIC COUNT: Manage automatic count through cache API.
      //   Check if we got some count in cache for
      //   'external_entities_<entity_type>_rest_count' and if so, return it.
      //   Note: if there are filtering parameters, we can't manage them for
      //   counting so we just ignore them and return the total count without
      //   filters so we may have additional empty pages but at least all
      //   results.
      return $this->configuration['endpoint_options']['count'];
    }
    $query_parameters = $this->toQueryParameters($parameters);

    $endpoint =
      $this->configuration['endpoint_options']['count']
      ?: $this->configuration['endpoint'];
    $method = $this->getRequestMethod('count', $endpoint);
    try {
      $req_parameters = NestedArray::mergeDeep(
        $this->getCountBodyParameters($query_parameters),
        [
          'headers' => $this->getHttpHeaders(),
          'query' => $this->getCountQueryParameters($query_parameters),
        ]
      );
      $this->ensureQueryLimits();
      $response = $this->httpClient->request(
        $method,
        $endpoint,
        $req_parameters
      );
    }
    catch (GuzzleException $exception) {
      $logger = $this->loggerChannelFactory->get('external_entities');
      $logger->error('Error in Rest::countQuery. Message: ' . $exception->getMessage());
      return 0;
    }

    $body = $this->getResponseBody(
      $response,
      'count',
      $method,
      $endpoint,
      $req_parameters
    );

    try {
      $results = $this
        ->getResponseDecoderFactory()
        ->getDecoder($this->configuration['response_format'])
        ->decode($body);
    }
    catch (InvalidDataTypeException $exception) {
      $logger = $this->loggerChannelFactory->get('external_entities');
      $logger->error('Error in decoding result. Message: ' . $exception->getMessage());
      return 0;
    }

    // Check if we must use a JSON Path to get data.
    if (!empty($this->configuration['data_path']['count'])) {
      try {
        $json_object = new JsonObject($results, TRUE);
        $results = $json_object->get($this->configuration['data_path']['count']);
        if (FALSE === $results) {
          $result = [];
        }
      }
      catch (InvalidJsonException | InvalidJsonPathException $e) {
        $logger = $this->loggerChannelFactory->get('external_entities');
        $logger->error(
          'Failed to parse JSON Path for count ('
          . $this->configuration['data_path']['count']
          . '): '
          . $e
        );
      }
    }

    // Manage array case.
    if (is_array($results)) {
      // Get first value of an array of a single numeric value.
      if ((1 == count($results))
          && (!empty($results[0]))
          && is_numeric($results[0])
      ) {
        $results = $results[0];
      }
      else {
        // Otherwise, assume we got an array of items to count.
        $results = count($results);
      }
    }

    // Check if the given value is numeric or a link including a count/end page.
    if (!is_numeric($results)) {
      // Check if we got a URL with a page/start item parameter.
      $page_param = $this->configuration['pager']['page_parameter'];
      if (!empty($page_param)
          && preg_match("#^(?:http|/).*[?&;]$page_param=(\d+)#i", $results, $matches)
      ) {
        $results = $matches[1];
      }
      elseif (preg_match('#(\d+)\D*$#', $results, $matches)) {
        // Otherwise, try to catch the last number.
        $results = $matches[1];
      }
      else {
        // Otherwise, we don't have a number.
        $results = 0;
      }
    }

    // Check count type.
    switch ($this->configuration['endpoint_options']['count_mode']) {
      case 'pages':
        // Page mode, multiply number by page size.
        $page_size =
          $this->configuration['pager']['default_limit']
          ?? static::DEFAULT_PAGE_LENGTH;
        $results *= $page_size;
        break;

      case 'entities':
      default:
        break;
    }

    return $results;
  }

  /**
   * Convert Drupal filter parameters to qeury parameters.
   *
   * @param array $parameters
   *   An array of filter parameters of the form:
   *   @code
   *     [
   *       [
   *         'field' => '<field_name>',
   *         'operator' => '=',
   *         'value' => '<field_value>',
   *       ],
   *       ...
   *     ]
   *   @endcode
   *
   * @return array
   *   An array with query parameter name as key and a string as their
   *   associated values.
   */
  protected function toQueryParameters(array $parameters) :array {
    $query_parameters = [];
    foreach ($parameters as $parameter) {
      if (!isset($parameter['field'])
        || !isset($parameter['value'])
        || (!empty($parameter['operator']) && ('=' != $parameter['operator']))
      ) {
        $logger = $this->loggerChannelFactory->get('external_entities');
        $logger->warning(
          'Unsupported filter parameter (skipped): '
          . print_r($parameter, TRUE)
        );
        continue;
      }
      $query_parameters[$parameter['field']] = $parameter['value'];
    }
    return $query_parameters;
  }

  /**
   * {@inheritdoc}
   *
   * For children classes, it is possible to override this method to support
   * other operators than '=' if possible and needed. For instance, if the
   * endpoint supports the 'STARTS_WITH' operator for the endpoint field
   * "blabla" using the query parameter "blablastartswith=", and that field
   * "blabla" is mapped to a Drupal field "field_bla", the override would remove
   * from $parameters filters that use "blabla" with the operator 'STARTS_WITH'
   * before passing it to parent::transliterateDrupalFilters() to map the rest
   * of parameters and add to the returned parameters the parameter
   * "blablastartswith" with the given value. In terms of code:
   * @code
   *   // child::transliterateDrupalFilters() input $parameters:
   *   [
   *     [
   *       'field' => 'field_bla',
   *       'value' => 'value_1',
   *       'operator' => 'STARTS_WITH',
   *     ],
   *     [
   *       'field' => 'field_other',
   *       'value' => ['value_2', 'value_3'],
   *       'operator' => 'IN',
   *     ],
   *     [
   *       'field' => 'field_yetanother',
   *       'value' => 'value_4',
   *       'operator' => '=',
   *     ],
   *     [
   *       'field' => 'field_op_not_supported',
   *       'value' => ['value_5', 'value_6'],
   *       'operator' => 'BETWEEN',
   *     ],
   *   ]
   *   // The child class could call Rest::transliterateDrupalFilters()
   *   // with new input $parameters:
   *   [
   *     [
   *       'field' => 'field_other',
   *       'value' => ['value_2', 'value_3'],
   *       'operator' => 'IN',
   *     ],
   *     [
   *       'field' => 'field_yetanother',
   *       'value' => 'value_4',
   *       'operator' => '=',
   *     ],
   *     [
   *       'field' => 'field_op_not_supported',
   *       'value' => ['value_5', 'value_6'],
   *       'operator' => 'BETWEEN',
   *     ],
   *   ]
   *   // The child class would get in return (assuming Drupal field
   *   // "field_other" is mapped to "other_source_name" and "field_yetanother"
   *   // to "yaf"):
   *   [
   *     'source' => [
   *       [
   *         'field' => 'other_source_name',
   *         'value' => 'value_2,value_3',
   *         'operator' => '=',
   *       ],
   *       [
   *         'field' => 'yaf',
   *         'value' => 'value_4',
   *         'operator' => '=',
   *       ],
   *     ],
   *     'drupal' => [
   *       [
   *         'field' => 'field_op_not_supported',
   *         'value' => ['value_5', 'value_6'],
   *         'operator' => 'BETWEEN',
   *       ],
   *     ],
   *   ]
   *   // That will be merged to the mapped "blabla" field and the child class
   *   // would return:
   *   [
   *     'source' => [
   *       [
   *         'field' => 'blablastartswith',
   *         'value' => 'value_1',
   *         'operator' => '=',
   *       ],
   *       [
   *         'field' => 'other_source_name',
   *         'value' => 'value_2,value_3',
   *         'operator' => '=',
   *       ],
   *       [
   *         'field' => 'yaf',
   *         'value' => 'value_4',
   *         'operator' => '=',
   *       ],
   *     ],
   *     'drupal' => [
   *       [
   *         'field' => 'field_op_not_supported',
   *         'value' => ['value_5', 'value_6'],
   *         'operator' => 'BETWEEN',
   *       ],
   *     ],
   *   ]
   *   // Note: since 'BETWEEN' operator is not supported by source, we don't
   *   // care which field 'field_op_not_supported' is mapped to since it will
   *   // be handled on the Drupal side using its own name.
   * @endcode
   * Note: a child class can transliterate a filter element into multiple
   * parameters if needed and vice-versa.
   * For an exemple of implementation, see
   * JsonAPI::transliterateDrupalFilters().
   * @see \Drupal\external_entities\Plugin\ExternalEntities\StorageClient\JsonAPI::transliterateDrupalFilters()
   *
   * @param array $parameters
   *   (optional) An array of Drupal-side field name parameter filtering like:
   *   @code
   *   [
   *     [
   *       'field' => 'field_name_1',
   *       'value' => 'value_1',
   *       'operator' => '=',
   *     ],
   *     [
   *       'field' => 'field_name_2',
   *       'value' => ['value_2', 'value_3',],
   *       'operator' => 'IN',
   *     ],
   *     ...
   *   ]
   *   @endcode
   *
   * @return array
   *   A 2 keys array with transliterated and not transliterated parameters
   *   array stored respectively under the keys 'source' and 'drupal'.
   */
  public function transliterateDrupalFilters(array $parameters) :array {
    // If no external entity is available, no way to map fields.
    if (empty($this->externalEntityType)) {
      return [
        'source' => [],
        'drupal' => $parameters,
      ];
    }
    $field_mapper = $this->externalEntityType->getFieldMapper();

    $source_filters = [];
    $drupal_filters = [];
    foreach ($parameters as $parameter) {
      if (!isset($parameter['field']) || !isset($parameter['value'])) {
        $drupal_filters[] = $parameter;
        continue;
      }

      $source_field = $field_mapper->getMappedFieldName($parameter['field']);
      if (!isset($source_field)) {
        $drupal_filters[] = $parameter;
        continue;
      }

      if (!empty($parameter['operator'])
          && (!in_array($parameter['operator'], ['=', 'IN']))
      ) {
        $drupal_filters[] = $parameter;
        $logger = $this->loggerChannelFactory->get('external_entities');
        $logger->warning(
          'Unsupported filter operator (skipped): '
          . print_r($parameter, TRUE)
        );
        continue;
      }
      if (is_array($parameter['value'])) {
        // @todo Maybe add a config form parameter to choose between imploding
        // with comas and using multiple time the query parameter name with
        // different values?
        $parameter['value'] = implode(',', $parameter['value']);
      }

      $parameter['field'] = $source_field;
      $parameter['operator'] = '=';
      $source_filters[] = $parameter;
    }

    return ['source' => $source_filters, 'drupal' => $drupal_filters];
  }

  /**
   * Returns the HTTP method to use for the given endpoint and request type.
   *
   * @param string $request_type
   *   The type of request. Should be one of create, read, update, delete,
   *   list or count.
   * @param string $endpoint
   *   Endpoint URL used for the call.
   *
   * @return string
   *   The HTTP method to use.
   */
  protected function getRequestMethod(
    string $request_type,
    string $endpoint
  ) :string {
    // Default ot GET.
    $method = 'GET';
    switch ($request_type) {
      case 'create':
        $method = 'POST';
        break;

      case 'read':
        break;

      case 'update':
        $method = 'PUT';
        break;

      case 'delete':
        $method = 'DELETE';
        break;

      case 'query':
        // Use POST for both 'body' and 'body_query' modes.
        if (0 === strpos($this->configuration['parameters']['list_param_mode'], 'body')) {
          $method = 'POST';
        }
        break;

      case 'count':
        // Will not use POST for count if set to 'body_query'.
        if ('body' == $this->configuration['parameters']['list_param_mode']) {
          $method = 'POST';
        }
        break;

      default:
        // We don't log unsupported request types as they might be handled by
        // extensions.s.
    }

    return $method;
  }

  /**
   * Returns the raw body content of the response.
   *
   * This method can be used by extensiosn to alter the body when needed or to
   * handle specific response codes like 202, 204 or 206 for instance.
   *
   * @param \Psr\Http\Message\ResponseInterface $response
   *   The HTTP response object.
   * @param string $request_type
   *   The type of request. Should be one of create, read, update, delete,
   *   list or count.
   * @param string $method
   *   HTTP request method to use. Could be one of 'GET', 'POST', 'PUT',
   *   'DELETE', 'PATCH', etc. See also getRequestMethod().
   * @param string $endpoint
   *   Endpoint URL used for the call.
   * @param array $parameters
   *   Array of parameters/options used for the request.
   *
   * @return string
   *   The raw body in a string, ready to be parsed.
   */
  protected function getResponseBody(
    ResponseInterface $response,
    string $request_type,
    string $method,
    string $endpoint,
    array $parameters = []
  ) :string {
    if (200 == $response->getStatusCode()) {
      // Force string.
      $body = $response->getBody() . '';
      // Remove UTF8 BOM added by some server which may intefer with data
      // parsers.
      $body = preg_replace('/^\xEF\xBB\xBF/', '', $body);
    }
    else {
      // Don't know how to handle non-200 answers: return an empty string.
      $body = '';
    }

    return $body;
  }

  /**
   * Adds global query parameters to the given set of parameters.
   *
   * @param array $parameters
   *   Specific parameters to use.
   *
   * @return array
   *   Query parameters.
   */
  public function getQueryParameters(array $parameters = []) {
    // Add authentication key to parameters if needed.
    if (!empty($this->configuration['api_key']['type'])
        && ('query' == $this->configuration['api_key']['type'])
        && !empty($this->configuration['api_key']['header_name'])
    ) {
      $parameters[$this->configuration['api_key']['header_name']] = $this->configuration['api_key']['key'];
    }

    return $parameters;
  }

  /**
   * Resturns request parameter structure related to body content.
   *
   * @param array $parameters
   *   Specific parameters to use.
   *
   * @return array
   *   An array with that should contain a key setting the type of body (one of
   *   'body', 'json', 'form_params' or 'multipart'), the adapted parameter
   *   value structure and an eventual 'header' key setting the content type or
   *   an empty array if nothing needed.
   */
  public function getBodyParameters(array $parameters = []) {
    if (empty($parameters)) {
      return [];
    }

    // Set mode.
    switch ($this->configuration['response_format']) {
      case 'json':
        // The content type "application/json" is automatically set by Guzzle.
        $parameters = ['json' => $parameters];
        break;

      case 'serialize':
        // PHP serialized data.
        $parameters = [
          'headers' => ['Content-Type' => 'application/vnd.php.serialized'],
          'body' => serialize($parameters),
        ];
        break;

      case 'yml':
        $parameters = [
          'headers' => ['Content-Type' => 'application/x-yaml'],
          'body' => Yaml::dump($parameters),
        ];
        break;

      // XML data, not supported at this time.
      // @code
      // case 'xml':
      //   $parameters = [
      //     'headers' => ['Content-Type' => 'application/xml'],
      //     'body' => ...($parameters),
      //   ];
      //   break;
      // @endcode
      //
      // Note for extensions: if you need to use files, use the body type
      // "'multipart' => $parameters" that will set content type to
      // "multipart/form-data".
      // @see https://docs.guzzlephp.org/en/stable/request-options.html#multipart
      default:
        // Guzzle will set content type to "application/x-www-form-urlencoded".
        $parameters = ['form_params' => $parameters];
    }

    return $parameters;
  }

  /**
   * Prepares and returns parameters used for list queries.
   *
   * This method is called by querySource() and countQuerySource(). The provided
   * $parameters parameter is supposed to contain only storage source compatible
   * parameters obtained from self::transliterateDrupalFilters().
   * This method will then add optional config-provided parameters (
   * authentication and custom parameters). The returned array will hold query
   * parameter name as keys and their associated values as strings (array values
   * are imploded with comas into strings).
   *
   * @param array $parameters
   *   (optional) An array of field filtering parameters like:
   *   @code
   *   [
   *     [
   *       'field' => 'name_1',
   *       'value' => 'value_1',
   *       'operator' => '=',
   *     ],
   *     [
   *       'field' => 'name_2',
   *       'value' => ['value_2', 'value_3', ],
   *       'operator' => 'IN',
   *     ],
   *     [
   *       'field' => 'name_3',
   *       'value' => 'value_4',
   *       'operator' => '<',
   *     ],
   *     ...
   *   ]
   *   @endcode
   *
   * @return array
   *   An associative array of query parameters keyed by parameter names with
   *   their associated values.
   *   Ex.:
   *   @code
   *   [
   *     'name_1' => 'value_1',
   *     'name_2' => 'value_2,value_3',
   *   ]
   *   @endcode
   *   Note: a child class could also implement the transliteration of the third
   *   example field if the source supports it that way, like this for instance:
   *   @code
   *     'name_3-lighter-than' => 'value_4',
   *   @endcode
   *   or even (it really depends how the source storage works and what
   *   self::transliterateDrupalFilters() implementation returned)
   *   @code
   *     'filter-1-field' => 'name_3',
   *     'filter-1-op'    => '>',
   *     'filter-1-value' => 'value_4',
   *   @endcode
   */
  public function getListQueryParameters(array $parameters = []) {
    $query_parameters = $this->toQueryParameters($parameters);

    if ((!empty($this->configuration['parameters']['list']))
      && ('query' == $this->configuration['parameters']['list_param_mode'])
    ) {
      $query_parameters += $this->configuration['parameters']['list'];
    }

    // Get query string from endpoint URL if specified.
    $endpoint = strstr($this->configuration['endpoint'], '?');
    if ($endpoint) {
      // Note: we don't consider ';' as a query parameter separator.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', substr($endpoint, 1), $matches);
      $query_parameters += array_combine($matches[1], $matches[2]);
    }

    return $this->getQueryParameters($query_parameters);
  }

  /**
   * Prepares and returns parameters used for count queries.
   *
   * @param array $parameters
   *   (optional) An array of Drupal-side field name parameter filtering. See
   *   getListQueryParameters() $parameters parameter for details.
   *
   * @return array
   *   An associative array of parameters.
   */
  public function getCountQueryParameters(array $parameters = []) {
    $query_parameters = $this->toQueryParameters($parameters);

    if ((!empty($this->configuration['parameters']['list']))
      && (FALSE !== strpos($this->configuration['parameters']['list_param_mode'], 'query'))
    ) {
      $query_parameters += $this->configuration['parameters']['list'];
    }

    // Get query string from endpoint URL if specified.
    $endpoint =
      $this->configuration['endpoint_options']['count']
      ?: $this->configuration['endpoint'];
    $endpoint = strstr($endpoint, '?');
    if ($endpoint) {
      // Note: we don't consider ';' as a query parameter separator.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', substr($endpoint, 1), $matches);
      $query_parameters += array_combine($matches[1], $matches[2]);
    }

    return $this->getQueryParameters($query_parameters);
  }

  /**
   * Returns request body parameters to use for list.
   *
   * @param array $parameters
   *   (optional) An array of Drupal-side field name parameter filtering. See
   *   getListQueryParameters() $parameters parameter for details.
   *
   * @return array
   *   See getBodyParameters().
   */
  public function getListBodyParameters(array $parameters = []) {
    $body_parameters = $this->toQueryParameters($parameters);

    // Use list parameters if mode set to 'body' or 'body_query'.
    if ((!empty($this->configuration['parameters']['list']))
      && (0 === strpos($this->configuration['parameters']['list_param_mode'], 'body'))
    ) {
      $body_parameters += $this->configuration['parameters']['list'];
    }

    return $this->getBodyParameters($body_parameters);
  }

  /**
   * Returns request body parameters to use for count.
   *
   * @param array $parameters
   *   (optional) An array of Drupal-side field name parameter filtering. See
   *   getListQueryParameters() $parameters parameter for details.
   *
   * @return array
   *   See getBodyParameters().
   */
  public function getCountBodyParameters(array $parameters = []) {
    $body_parameters = $this->toQueryParameters($parameters);

    // Do not use list parameters if mode set to 'query' or 'body_query'.
    if ((!empty($this->configuration['parameters']['list']))
      && ('body' == $this->configuration['parameters']['list_param_mode'])
    ) {
      $body_parameters += $this->configuration['parameters']['list'];
    }

    return $this->getBodyParameters($body_parameters);
  }

  /**
   * Gets the paging query parameters based on the configuration.
   *
   * Since the number of requested entities may be higher than the maximum
   * number of entities supported by a single endpoint listing call, this
   * function will return an array of paging parameters for subsequent requests
   * to list all the requested entities.
   *
   * Also, since the REST service imposes a page size, we will have to deal with
   * it. For instance, consider a set of 10 entities with a REST service
   * allowing a maximum page size of 4 entities and using a page number
   * parameter. If the query requests 3 entities starting from 2  (zero-based
   * indexing), we will have to query the service 2 times. The first call will
   * get page 0 with entities 0 to 3 and the second call will get entities from
   * 4 to 7. Doing so, we will get a set of 8 entities from 0 to 7 while the
   * query was only expecting 3 entities starting from 2. Therefore, this method
   * will return an array with 2 sub-arrays. Each sub-array will hold the query
   * parameters in a 'parameters' key and the entities to keep in the query
   * result set using the 'start' en 'length' keys. The method caller will have
   * the responsability to filter the requested entities.
   * Example of the resulting array:
   * @code
   *   [
   *     [
   *       'parameters' => ['page' => 0, 'pagesize' => 4, ],
   *       'start' => 2,
   *       'length' => 2,
   *     ],
   *     [
   *       'parameters' => ['page' => 1, 'pagesize' => 4, ],
   *       'start' => 0,
   *       'length' => 1,
   *     ],
   *   ];
   * @endcode
   * But this methods also handles the use of starting item instead of page
   * number and ending item instead of number of items per page. The returned
   * arrays will have the same structure. If the previous example used a start
   * item and an ending item parameters with the same page size limitation, it
   * would look like this:
   * @code
   *   [
   *     [
   *       'parameters' => ['from' => 2, 'to' => 4, ],
   *       'start' => 0,
   *       'length' => 3,
   *     ],
   *   ];
   * @endcode
   * However, if it requested 5 elements instead of 3, it would look like this:
   * @code
   *   [
   *     [
   *       'parameters' => ['from' => 2, 'to' => 5, ],
   *       'start' => 0,
   *       'length' => 4,
   *     ],
   *     [
   *       'parameters' => ['from' => 6, 'to' => 6, ],
   *       'start' => 0,
   *       'length' => 1,
   *     ],
   *   ];
   * @endcode
   *
   * @param int $start
   *   (optional) Item index to start with.
   * @param int $length
   *   (optional) Amount of items to return.
   *
   * @return array
   *   An array of paging parameters. An empty array if no paging is available.
   *   Each paging parameter contains a key 'parameters' with an associative
   *   array of query parameters for paging, a key 'start' containing the index
   *   of the first element to keep from the query and a key 'length' indicating
   *   how many elements to keep from the query.
   */
  public function getPagingQueryParameters(?int $start = NULL, ?int $length = NULL) :array {
    $paging_parameters = [];
    // Check parameters.
    $start ??= 0;
    if (0 > $start) {
      $start = 0;
    }
    if (!isset($length) || (0 > $length)) {
      // Set length to 0 (empty value) to be set later.
      $length = 0;
    }
    // Only apply paging if we have the parameters for it.
    if (!empty($this->configuration['pager']['page_parameter'])
        && !empty($this->configuration['pager']['page_size_parameter'])
    ) {
      $one_offset = empty($this->configuration['pager']['page_start_one'])
        ? 0
        : 1;
      $end_item_mode = ($this->configuration['pager']['page_size_parameter_type'] === 'enditem');
      $pagenum_mode = ($this->configuration['pager']['page_parameter_type'] === 'pagenum');

      $max_page_size = ($this->configuration['pager']['default_limit'] ?? 0)
        ?: static::DEFAULT_PAGE_LENGTH;
      $length = $length ?: $max_page_size;
      $page_size = min($max_page_size, ($start + $length));
      $keep_start = 0;
      $keep_length = $page_size;
      $ending_pos = $start + $length - 1;

      if ($pagenum_mode) {
        // Convert start item into a page number.
        $keep_start = $start % $max_page_size;
        $start = (int) floor($start / $max_page_size);
        $keep_length = min(
          $page_size - $keep_start,
          $ending_pos + 1 - ($start * $max_page_size)
        );
        if ($end_item_mode) {
          $page_size = min($ending_pos, ($start + 1) * $max_page_size - 1) + $one_offset;
        }
      }
      else {
        // Start item mode.
        $keep_length = min($length, $max_page_size);
        if ($end_item_mode) {
          $page_size = min($ending_pos, $start + $max_page_size - 1) + $one_offset;
        }
      }

      while (0 < $keep_length) {
        $paging_parameters[] = [
          'parameters' => [
            $this->configuration['pager']['page_parameter'] => $start + $one_offset,
            $this->configuration['pager']['page_size_parameter'] => $page_size,
          ],
          'start' => $keep_start,
          'length' => $keep_length,
        ];

        // Compute next paging data...
        $keep_start = 0;
        if ($pagenum_mode) {
          // Page number, got to next page.
          ++$start;
          // Get current page end position.
          $current_page_end = (($start + 1) * $max_page_size) - 1;

          // No need to update $page_size for "number of items per page" mode.
          if ($end_item_mode) {
            // Update end item to either the final position or the last position
            // in current page.
            $page_size = min($ending_pos, $current_page_end) + $one_offset;
          }

          // Update $keep_length to $max_page_size unless we reached last page.
          $keep_length = min(
            $max_page_size,
            $ending_pos - $keep_start + 1 - ($start * $max_page_size)
          );
        }
        else {
          // Paging in starting item mode, go forward for $max_page_size items.
          $start += $max_page_size;

          // Update $page_size.
          if ($end_item_mode) {
            // Update end item to either the final position or the last position
            // in current page.
            $page_size = min($ending_pos, $start + $max_page_size - 1) + $one_offset;
          }
          else {
            // Update end item to either the final position or the maximum
            // number of item in a page.
            $page_size = min($ending_pos + 1 - $start, $max_page_size);
          }

          // Update $keep_length to $max_page_size unless we reached last page.
          $keep_length = min(
            $max_page_size,
            $ending_pos + 1 - $start
          );
        }
      }
    }
    else {
      // No paging parameters.
      // Set a default length.
      $length = $length ?: static::DEFAULT_PAGE_LENGTH;
      // Limit length if needed.
      if (!empty($this->configuration['pager']['default_limit'])) {
        $length = min($length, $this->configuration['pager']['default_limit'] - $start);
      }
      $paging_parameters = [
        [
          'parameters' => [],
          'start' => $start,
          'length' => $length,
        ],
      ];
    }
    return $paging_parameters;
  }

  /**
   * Prepares and returns parameters used for single item queries.
   *
   * This method is called by load() and will provide optional config-provided
   * parameters (authentication and custom parameters). "{id}" placeholders in
   * custom parameter values are replaced by the provided entity identifier
   * value ($id). The returned array will hold query parameter name as keys and
   * their associated values as strings.
   *
   * @param int|string|null $id
   *   The item id being fetched.
   *
   * @return array
   *   An associative array of parameters.
   */
  public function getSingleQueryParameters(int|string|null $id) :array {
    $query_parameters = [];

    // Get query string from endpoint URL if specified.
    $endpoint =
      $this->configuration['endpoint_options']['single']
      ?: $this->configuration['endpoint'];
    $endpoint = strstr($endpoint, '?');
    if ($endpoint) {
      // Note: we don't consider ';' as a query parameter separator.
      preg_match_all('#([^&=]+)(?:=([^&=]*))?#', substr($endpoint, 1), $matches);
      $query_parameters += array_combine($matches[1], $matches[2]);
    }

    if ((!empty($this->configuration['parameters']['single']))
      && ('query' == $this->configuration['parameters']['single_param_mode'])
    ) {
      if (!empty($id) || (0 === $id)) {
        $query_parameters += array_map(
          function ($value) use ($id) {
            return str_replace('{id}', $id ?? '', $value);
          },
          $this->configuration['parameters']['single']
        );
      }
      else {
        $query_parameters += $this->configuration['parameters']['single'];
      }
    }
    return $this->getQueryParameters($query_parameters);
  }

  /**
   * Returns body parameters to use for single entity request.
   *
   * @param int|string|null $id
   *   The item id being fetched.
   * @param array $data
   *   Extra data values to add to the body. For instance, entity field values
   *   when saving an entity.
   *
   * @return array
   *   An associative array of parameters.
   */
  public function getSingleBodyParameters(int|string|null $id, array $data = []) :array {
    $body_data = [];

    if (!empty($this->configuration['parameters']['single'])
      && ('body' == $this->configuration['parameters']['single_param_mode'])
    ) {
      // Replace entity identifier if needed.
      if (!empty($id) || (0 === $id)) {
        $body_data += array_map(
          function ($value) use ($id) {
            return str_replace('{id}', $id ?? '', $value);
          },
          $this->configuration['parameters']['single']
        );
      }
      else {
        $body_data += $this->configuration['parameters']['single'];
      }
    }

    // Add extra data if needed.
    if (!empty($data)) {
      $body_data += $data;
    }

    return $this->getBodyParameters($body_data);
  }

  /**
   * Gets the HTTP headers to add to a request.
   *
   * @return array
   *   Associative array of headers to add to the request.
   */
  public function getHttpHeaders() :array {
    $headers = [];

    if (!empty($this->configuration['http']['headers'])) {
      $header_lines = explode("\n", $this->configuration['http']['headers']);
      foreach ($header_lines as $header_line) {
        @[$header_name, $value] = explode(':', $header_line, 2);
        $value ??= '';
        $headers[$header_name] = trim($value);
      }
    }

    if (!empty($this->configuration['api_key']['type'])
        && (in_array($this->configuration['api_key']['type'], ['bearer', 'custom']))
        && !empty($this->configuration['api_key']['header_name'])
        && !empty($this->configuration['api_key']['key'])
    ) {
      $headers[$this->configuration['api_key']['header_name']] = $this->configuration['api_key']['key'];
    }

    return $headers;
  }

  /**
   * Sleep a certain amount of time if needed.
   *
   * This method just waits the necessary amount of time before returning to let
   * the system query the endpoint. If the waiting time is longer than PHP
   * setting 'max_execution_time', it will not wait more so it may be possible
   * the endpoint got more queries than allowed in the given amount of time.
   * Also note that this method is far from perfect and can suffer from race
   * condition: two parallel calls may lead to one request time not being noted.
   */
  protected function ensureQueryLimits() {
    $qcount = $this->configuration['endpoint_options']['limit_qcount'];
    if (empty($qcount)) {
      return;
    }
    $qsec = $this->configuration['endpoint_options']['limit_qtime'];
    if (empty($qsec)) {
      return;
    }

    $endpoint = $this->configuration['endpoint'];

    $wait_limit = ini_get('max_execution_time') ?: 30;
    $wait_time = 0;
    while (TRUE) {
      $qtime = time();
      // Get query count for endpoint.
      $qtimes = $this->connection
        ->select('xntt_rest_queries', 'q')
        ->fields('q', ['qtimes'])
        ->condition('q.endpoint', $endpoint)
        ->execute()
        ->fetchAssoc()
        ?: ['qtimes' => ''];
      $queries = explode(',', $qtimes['qtimes']);

      // Cleanup old queries.
      $oldest = $qtime - $qsec;
      $need_update = FALSE;
      while (!empty($queries) && ($queries[0] < $oldest)) {
        array_shift($queries);
        $need_update = TRUE;
      }

      // Update db.
      if ($need_update) {
        $value = [
          'endpoint' => $endpoint,
          'qtimes' => implode(',', $queries),
          'xntt_type' => $this->externalEntityType->id(),
          'username' => \Drupal::currentUser()->getAccountName(),
        ];
        $this->connection
          ->upsert('xntt_rest_queries')
          ->fields(['endpoint', 'qtimes', 'xntt_type', 'username'])
          ->key('endpoint')
          ->values($value)
          ->execute();
      }

      ++$wait_time;
      // Check query count in the amount of time (and max wait time).
      if ((count($queries) < $qcount) || ($wait_time > $wait_limit)) {
        break;
      }
      // Wait 200ms.
      usleep(200000);
    }

    // Add current query time.
    $queries[] = time();
    $value = [
      'endpoint' => $endpoint,
      'qtimes' => implode(',', $queries),
      'xntt_type' => $this->externalEntityType->id(),
      'username' => \Drupal::currentUser()->getAccountName(),
    ];
    $this->connection
      ->upsert('xntt_rest_queries')
      ->fields(['endpoint', 'qtimes', 'xntt_type', 'username'])
      ->key('endpoint')
      ->values($value)
      ->execute();
  }

}
