<?php

namespace Drupal\external_entities\Plugin\ExternalEntities\StorageClient;

use Drupal\Core\Form\FormStateInterface;

/**
 * External entities storage client based on a REST API.
 *
 * @StorageClient(
 *   id = "wiki",
 *   label = @Translation("Wiki"),
 *   description = @Translation("Retrieves external entities from a Mediawiki API (Wikipedia).")
 * )
 */
class Wiki extends Rest {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form = parent::buildConfigurationForm($form, $form_state);

    // Adjust default REST form.
    $form['endpoint']['#title'] = $this->t('Wiki API URL');
    $form['endpoint']['#description'] = $this->t(
      'Should be something similar to "<code>https://www.example.org/w/api.php</code>".'
    );
    @$form['endpoint']['#default_value'] = $this->configuration['endpoint'] ?: "https://en.wikipedia.org/w/api.php";
    $form['endpoint']['#attributes']['#placeholder'] = $this->t("ex.: https://en.wikipedia.org/w/api.php");

    $form['endpoint_options']['cache'] = [
      '#type' => 'hidden',
      '#value' => FALSE,
    ];

    $form['endpoint_options']['single'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    $form['endpoint_options']['count'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Total number of wiki pages'),
      '#required' => FALSE,
      '#description' => $this->t('Only use this field if you are not using a "search" list type of query. Otherwise, "search" list type supports a "srinfo=totalhits" parameter that provides the number of maching pages.'),
      '#default_value' => $this->configuration['endpoint_options']['count'] ?? '',
    ];

    $form['endpoint_options']['count_mode'] = [
      '#type' => 'hidden',
      '#value' => 'entities',
    ];

    $form['endpoint_options']['limit_qcount'] = [
      '#type' => 'hidden',
      '#value' => '200',
    ];

    $form['endpoint_options']['limit_qtime'] = [
      '#type' => 'hidden',
      '#value' => '1',
    ];

    $form['response_format'] = [
      '#type' => 'hidden',
      '#value' => 'json',
    ];

    $form['data_path']['#open'] = FALSE;

    $form['data_path']['list'] = [
      '#type' => 'hidden',
      '#value' => '$.query.*.*',
    ];

    $form['data_path']['single'] = [
      '#type' => 'hidden',
      '#value' => '$.query.pages',
    ];

    $form['data_path']['keyed_by_id'] = [
      '#type' => 'hidden',
      '#value' => FALSE,
    ];

    // Do not warn if not set yet.
    @$form['data_path']['count']['#default_value'] = $this->configuration['data_path']['count'] ?: '$.query.searchinfo.totalhits';
    $form['data_path']['count']['#attributes']['placeholder'] = $this->t('ex.: $.query.searchinfo.totalhits');

    $form['pager']['#open'] = FALSE;

    // Pager limit is auto-set using '*limit' parameters.
    // Default Wiki limit is 10 but it's better for Drupal to use a 50 elements
    // basis (1 query/page).
    $form['pager']['default_limit'] = [
      '#type' => 'hidden',
      '#value' => static::DEFAULT_PAGE_LENGTH,
    ];

    $form['pager']['page_parameter']['#attributes']['placeholder'] = $this->t('ex.: sroffset');

    $form['pager']['page_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'startitem',
    ];

    $form['pager']['page_size_parameter']['#attributes']['placeholder'] = $this->t('ex.: srlimit');

    $form['pager']['page_size_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'pagesize',
    ];

    $form['pager']['page_start_one'] = [
      '#type' => 'hidden',
      '#value' => FALSE,
    ];

    unset($form['api_key']['#type']);

    $form['api_key']['type'] = [
      '#type' => 'hidden',
      '#value' => 'none',
    ];

    $form['api_key']['key'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    $form['api_key']['header_name'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    unset($form['http']['#type']);

    $form['http']['headers'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    $form['parameters']['#open'] = TRUE;

    $form['parameters']['list']['#attributes']['placeholder'] = $this->t("ex.:\naction|query\nformat|json\nlist|search\srsearch=Haarlem\nsrprop|categorysnippet\nsrinfo|totalhits");
    $form['parameters']['list_param_mode'] = [
      '#type' => 'hidden',
      '#value' => 'query',
    ];
    $form['parameters']['single']['#attributes']['placeholder'] = $this->t("ex.:\naction|query\nformat|json\nprop|extracts");
    $form['parameters']['single_param_mode'] = [
      '#type' => 'hidden',
      '#value' => 'query',
    ];

    $form['parameters']['default_params'] = [
      '#type' => 'markup',
      '#markup' => $this->t('<b>NOTE:</b> The following parameters are automatically set: <code>action|query</code> and <code>format|json</code>. You should provide at least a "list" parameter with its corresponding required arguments. It is better to use "<a href="https://www.mediawiki.org/wiki/API:Search">search</a>" list type which provides paging (<code>srlimit</code>, <code>sroffset</code> and <code>srinfo=totalhits</code>) otherwise, you will have hard time to simulate a correct paging or will be limited to one page of results. See <a href="https://www.mediawiki.org/wiki/API:Lists">MediaWiki List API documentation</a> for details.'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    $parameters = $form_state->getValue('parameters');
    foreach ($parameters as $type => &$value) {
      // Ensure query action and JSON response format are set.
      // First, clear settings for action and format.
      $value = preg_replace('#(?<=^|\n)(?:action|format)\|[^\n]*\n?#', '', $value);
      $value = "action|query\nformat|json\n" . $value;
    }
    $list_param = &$parameters['list'];

    // Check for list type.
    if (!preg_match('#(?:^|\n)list\|([-:\w]+)#', $parameters['list'], $list_matches)) {
      // Warn the single enpoint URL may not be correct.
      \Drupal::messenger()->addWarning('You should provide a "list|<list_type>" parameter in the "List parameters" section.');
    }
    elseif ('search' == $list_matches[1]) {
      // If 'search' type, make sure we got srinfo.
      $list_param = preg_replace('#(?<=^|\n)(?:srinfo)\|[^\n]*\n?#', '', $list_param);
      $list_param .= "\nsrinfo=totalhits";
    }
    $prefix = '';
    if (!empty($list_matches[1])) {
      $prefix = $this->getListTypePrefix($list_matches[1]);
    }

    // Check for pager config.
    $pager = $form_state->getValue('pager');
    if (preg_match('#(?:^|\n)\w{0,3}limit\|(\d+)#', $parameters['list'], $matches)) {
      $pager['default_limit'] = $matches[1];
    }
    elseif ($prefix) {
      // Add default limit (for Drupal).
      $list_param =
        trim($list_param, "\n")
        . "\n{$prefix}limit|"
        . static::DEFAULT_PAGE_LENGTH
        . "\n";
    }

    if ($prefix) {
      // Remove pager offset from parameters and add offset and limit to the
      // pager config.
      $list_param = preg_replace(
        '#(?<=^|\n)' . $prefix . 'offset\|[^\n]*\n?#',
        '',
        $list_param
      );
      if (empty($pager['page_parameter'])) {
        $pager['page_parameter'] = $prefix . 'offset';
      }
      if (empty($pager['page_size_parameter'])) {
        $pager['page_size_parameter'] = $prefix . 'limit';
      }
    }

    $form_state->setValue('parameters', $parameters);
    $form_state->setValue('pager', $pager);
    parent::validateConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function getSingleQueryParameters($id, array $parameters = []) :array {
    return parent::getSingleQueryParameters($id, [
      [
        'field' => 'pageids',
        'value' => $id,
      ],
    ]);
  }

  /**
   * Returns Wikipedia (API) prefix for the given list type.
   *
   * @param string $list_type
   *   The list type. Ex.: 'categorymembers'.
   *
   * @return string:
   *   The prefix used in the API or an empty string if not available.
   */
  public function getListTypePrefix(string $list_type) :string {
    static $prefixes = [
      'search'              => 'sr',
      'allcategories'       => 'ac',
      'alldeletedrevisions' => 'adr',
      'allfileusages'       => 'af',
      'allimages'           => 'ai',
      'alllinks'            => 'al',
      'allpages'            => 'ap',
      'allredirects'        => 'ar',
      'allrevisions'        => 'arv',
      'alltransclusions'    => 'at',
      'allusers'            => 'au',
      'backlinks'           => 'bl',
      'blocks'              => 'bk',
      'categorymembers'     => 'cm',
      'deletedrevs'         => 'dr',
      'embeddedin'          => 'ei',
      'exturlusage'         => 'eu',
      'filearchive'         => 'fa',
      'imageusage'          => 'iu',
      'iwbacklinks'         => 'iwbl',
      'langbacklinks'       => 'lbl',
      'logevents'           => 'le',
      'mystashedfiles'      => 'msf',
      'pagepropnames'       => 'ppn',
      'pageswithprop'       => 'pwp',
      'prefixsearch'        => 'ps',
      'protectedtitles'     => 'pt',
      'querypage'           => 'qp',
      'random'              => 'rn',
      'recentchanges'       => 'rc',
      'search'              => 'sr',
      'tags'                => 'tg',
      'usercontribs'        => 'uc',
      'users'               => 'us',
      'watchlist'           => 'wl',
      'watchlistraw'        => 'wr',
    ];
    return $prefixes[$list_type] ?? '';
  }

}
