<?php

namespace Drupal\external_entities\StorageClient;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Plugin\PluginDependencyTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\external_entities\ExternalEntityTypeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for external entity storage clients.
 */
abstract class StorageClientBase extends PluginBase implements StorageClientInterface {

  // Normally, we'd just need \Drupal\Core\Entity\DependencyTrait here for
  // plugins. However, in a few cases, plugins use plugins themselves, and then
  // the additional self::calculatePluginDependencies() method from this trait
  // is useful. Since PHP 5 complains when adding this trait along with its
  // "parent" trait to the same class, we just add it here in case a child class
  // does need it.
  use PluginDependencyTrait;

  /**
   * Debug level of current instance.
   *
   * @var int
   */
  protected $debugLevel;

  /**
   * The external entity type this storage client is configured for.
   *
   * @var \Drupal\external_entities\ExternalEntityTypeInterface
   */
  protected $externalEntityType;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Constructs a StorageClientBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    EntityFieldManagerInterface $entity_field_manager
  ) {
    $configuration += $this->defaultConfiguration();
    if (!empty($configuration['_external_entity_type'])
        && $configuration['_external_entity_type'] instanceof ExternalEntityTypeInterface
    ) {
      $this->externalEntityType = $configuration['_external_entity_type'];
      unset($configuration['_external_entity_type']);
    }
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->setStringTranslation($string_translation);
    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('entity_field.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel() :string {
    $plugin_definition = $this->getPluginDefinition();
    return $plugin_definition['label'];
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() :string {
    $plugin_definition = $this->getPluginDefinition();
    return $plugin_definition['description'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getDebugLevel() :int {
    return $this->debugLevel ?? 0;
  }

  /**
   * {@inheritdoc}
   */
  public function setDebugLevel(int $debug_level = 1) {
    $this->debugLevel = $debug_level;
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = NestedArray::mergeDeep(
      $this->defaultConfiguration(),
      $configuration
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function load(string|int $id) :array|null {
    return current($this->loadMultiple([$id])) ?: NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuerySource(array $parameters = []) :int {
    // Default inefficient implementation.
    return count($this->querySource($parameters));
  }

  /**
   * {@inheritdoc}
   */
  public function query(
    array $parameters = [],
    array $sorts = [],
    ?int $start = NULL,
    ?int $length = NULL
  ) :array {
    // Translate filters.
    $trans_parameters = $this->transliterateDrupalFilters($parameters);

    // Get filtered results from source.
    if (empty($trans_parameters['drupal'])) {
      // We can use source paging and sorting.
      $results = $this->querySource(
        $trans_parameters['source'] ?? [],
        $sorts,
        $start,
        $length
      );
    }
    else {
      // Get all and process to a second pass.
      $first_pass_results = $this->querySource(
        $trans_parameters['source']
        ?? []
      );
      $results = $this->postFilterQuery(
        $first_pass_results,
        $trans_parameters['drupal']
      );
      // Process sorting.
      // @todo Manage field properties field names like "field_geolocation.lat".
      usort($results, function ($a, $b) use ($sorts) {
        foreach ($sorts as $sort) {
          if (($sort[1] ?? '') == 'DESC') {
            // Descending.
            $before = 1;
            $after = -1;
          }
          else {
            // Ascending.
            $before = -1;
            $after = 1;
          }
          if (array_key_exists($sort[0], $a)) {
            if (array_key_exists($sort[0], $b)) {
              if ($a[$sort[0]] != $b[$sort[0]]) {
                if (is_numeric($a[$sort[0]]) && is_numeric($b[$sort[0]])) {
                  // Numeric comparison.
                  return (($a - $b) < 0) ? $before : $after;
                }
                else {
                  // Text comparison.
                  return (strcmp($a[$sort[0]], $b[$sort[0]]) < 0)
                    ? $before
                    : $after;
                }
              }
              // Equality, continue next sort.
            }
            else {
              // $b does not have the field.
              return $after;
            }
          }
          else {
            // $a does not have the field.
            if (array_key_exists($sort[0], $b)) {
              // $b does.
              return $before;
            }
            // None have the field, continue next sort.
          }
        }
        // Equality.
        return 0;
      });
      // Process paging.
      if (isset($length)) {
        $results = array_splice($results, $start ?? 0, $length);
      }
      elseif (isset($start)) {
        $results = array_splice($results, $start);
      }
    }
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuery(array $parameters = []) :int {
    // Translate filters.
    $trans_parameters = $this->transliterateDrupalFilters($parameters);
    // Get filtered results from source.
    if (empty($trans_parameters['drupal'])) {
      // We can use source count.
      return $this->countQuerySource($trans_parameters['source'] ?? []);
    }
    else {
      // We need to count post-filtered entities.
      return count($this->query($parameters));
    }
  }

  /**
   * Filter external entitiy arrays according to the given filters.
   *
   * Method used to filter the given external entities using Drupal-type filters
   * provided in the $parameters array. This method is used to filter values
   * that can not be directly filtered on the data source side (storage side).
   *
   * @param array $result_set
   *   Array of external entity arrays.
   * @param array $parameters
   *   Array of filter parameters. Each parameter is an array like that:
   *   @code
   *     [
   *       'field' => <external entity Drupal field name (string)>,
   *       'value' => <filter value, may be NULL (mixed)>,
   *       'operator' => <an operator (string)>,
   *     ]
   *   @endcode
   *   See self::testDrupalFilter() for possible operators.
   *
   * @return array
   *   An array of external entities passing all the filters or an empty array
   *   if none did.
   */
  protected function postFilterQuery(
    array $result_set,
    array $parameters
  ) :array {
    $results = [];
    // Process each entity array.
    foreach ($result_set as $result) {
      // Process remaining filters.
      foreach ($parameters as $filter) {
        $field_value = $result[$filter['field']] ?? NULL;
        $operator = $filter['operator'];
        $filter_value = $filter['value'] ?? NULL;
        if (!$this->testDrupalFilter($field_value, $filter_value, $operator)) {
          // Test failed, filter not passed, skip entity and process next.
          continue 2;
        }
      }
      // Entity passed all filters, keep.
      $results[] = $result;
    }
    return $results;
  }

  /**
   * Returns TRUE if the given field value passes the given filter.
   *
   * @param mixed $field_value
   *   The field value to test.
   * @param mixed $filter_value
   *   The optional filter value which may be a numeric value, a string, NULL
   *   for "IS (NOT) NULL/(NOT) EXISTS" operators, or an array of 2 numeric
   *   value for "(NOT) BETWEEN" operator, or an array of values for "(NOT) IN"
   *   operator.
   * @param string $operator
   *   The operator to use which should be one of '=', '!=', '<>', '>', '>=',
   *   '<', '<=', 'STARTS_WITH', 'CONTAINS', 'ENDS_WITH', 'IN', 'NOT IN',
   *   'IS NULL', 'IS NOT NULL', 'EXISTS', 'NOT EXISTS', 'BETWEEN' and
   *   'NOT BETWEEN'.
   *   Note: '!=', 'EXISTS' and 'NOT EXISTS' are not usually used by Drupal and
   *   only supported for convenience.
   *
   * @return bool
   *   TRUE if the given field values matches the given filter value according
   *   to the given filter if supported, FALSE in any other cases.
   */
  protected function testDrupalFilter(
    $field_value,
    $filter_value,
    string $operator
  ) :bool {
    // @todo Add event to let plugins support other filters?
    // @todo Manage field properties field names like "field_geolocation.lat".
    $pass = FALSE;
    switch ($operator) {
      case '=':
        $pass = ($field_value == $filter_value);
        break;

      case '<>':
      case '!=':
        $pass = ($field_value != $filter_value);
        break;

      case '>':
        $pass = ($field_value > $filter_value);
        break;

      case '>=':
        $pass = ($field_value >= $filter_value);
        break;

      case '<':
        $pass = ($field_value < $filter_value);
        break;

      case '<=':
        $pass = ($field_value <= $filter_value);
        break;

      case 'STARTS_WITH':
        $pass =
          (substr($field_value, 0, strlen($filter_value)) == $filter_value);
        break;

      case 'CONTAINS':
        $pass = (FALSE !== strpos($field_value, $filter_value));
        break;

      case 'ENDS_WITH':
        $pass =
          (substr($field_value, -1 * strlen($filter_value)) == $filter_value);
        break;

      case 'IN':
        if (is_array($filter_value)) {
          $pass = in_array($field_value, $filter_value);
        }

      case 'NOT IN':
        if (is_array($filter_value)) {
          $pass = !in_array($field_value, $filter_value);
        }

      case 'IS NULL':
      case 'NOT EXISTS':
        $pass = !isset($field_value);
        break;

      case 'IS NOT NULL':
      case 'EXISTS':
        $pass = isset($field_value);
        break;

      case 'BETWEEN':
        if (is_numeric($field_value)
            && is_array($filter_value)
            && (2 == count($filter_value))
            && is_numeric($filter_value[0])
            && is_numeric($filter_value[1])
        ) {
          $pass = ($filter_value[0] <= $field_value)
            && ($field_value <= $filter_value[1]
          );
        }

      case 'NOT BETWEEN':
        if (is_numeric($field_value)
            && is_array($filter_value)
            && (2 == count($filter_value))
            && is_numeric($filter_value[0])
            && is_numeric($filter_value[1])
        ) {
          $pass = ($field_value < $filter_value[0])
            || ($filter_value[1] < $field_value
          );
        }

      default:
        break;
    }

    return $pass;
  }

  /**
   * Returns the list of field definition this storage client external entity.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface[]
   *   An array of field definitions.
   */
  protected function getFieldDefinitions() :array {
    if (empty($this->externalEntityType)) {
      return [];
    }

    // Get field definitions.
    $xntt_type = $this->externalEntityType->id();
    $field_defs = $this
      ->entityFieldManager
      ->getFieldDefinitions($xntt_type, $xntt_type);
    return $field_defs;
  }

  /**
   * Returns the related external entity field mapping.
   *
   * @return array
   *   Field mapping as returned by FieldMapperInterface::getFieldMappings().
   *
   * @see \Drupal\external_entities\FieldMapper\FieldMapperInterface::getFieldMappings()
   */
  protected function getFieldMappings() :array {
    if (empty($this->externalEntityType)) {
      return [];
    }

    return $this->externalEntityType->getFieldMapper()->getFieldMappings();
  }

  /**
   * Returns the source field name mapped to Drupal external entity identifier.
   *
   * @return string|null
   *   The source field name used to store identifier or NULL if not available.
   */
  public function getSourceIdFieldName() :?string {
    if (empty($this->externalEntityType)) {
      return NULL;
    }
    return $this
      ->externalEntityType
      ->getFieldMapper()
      ->getMappedFieldName('id', 'value');
  }

}
